/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"

static real rho, rhoLast, alpha, beta, sigma;
static int cgsu = -1, cgsv = -1, cgsp = -1, cgsq = -1, cgsr = -1, 
cgsr0 = -1, cgsh1 = -1, cgsh2 = -1;

static int CGSInit(iteMethod *itor) 
  {
    char s[256];
    matMulMethod *matMul = itor->matMul;
    preCondMethod *preCond = itor->preCond;

    if (matMul==nil) matMul = actMatMul;
    if (preCond==nil) preCond = actPreCond;
    if (cgsu==-1)
      cgsu = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsu");
    if (cgsv==-1)
      cgsv = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsv");
    if (cgsp==-1)
      cgsp = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsp");
    if (cgsq==-1)
      cgsq = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsq");
    if (cgsr==-1)
      cgsr = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsr");
    if (cgsh1==-1)
      cgsh1 = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsh1");
    if (cgsh2==-1)
      cgsh2 = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsh2");
    if (cgsr0==-1)
      cgsr0 = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"cgsr0");
    if (cgsr0==-1) return false;

    if (itor->verboseP)
      {
        sprintf(s, "Solve: start of CGS, n=%d, eps=%9.3e, steps=(%d,%d) %s\n", 
                GetNoNodes(), itor->eps, itor->minSteps, itor->maxSteps,
                preCond->name);
        ZIBStdOut(s);
      }
    
    if (matMul->InitMatMul)
      (matMul->InitMatMul)(-1,actProblem->LocAss,nodesState->rRhs, 
                           actProblem->LocAssB);
    if (preCond->InitPreCond) (preCond->InitPreCond)(); 
    
    if ((matMul->MatMul)(nodesState->rSol, cgsr)==false) return false;
    NodeSub(cgsr,nodesState->rRhs,cgsr);
    NodeAssign(cgsr0,cgsr);
    NodeAssign(cgsp,cgsr);
    NodeAssign(cgsu,cgsr);
    if (preCond->PreCond) (preCond->PreCond)(cgsp,cgsh1); 
     else NodeAssign(cgsh1,cgsp);
    if ((matMul->MatMul)(cgsh1,cgsv)==false) return false;
    rho = NodeScalProd(cgsr0,cgsr0);
    sigma = NodeScalProd(cgsv,cgsr0);

    itor->iteCount = 0;
    
    return true;
  }

static int CGSIterate(iteMethod *itor) 
  {
    int it = itor->iteCount;
    real no = GetNoNodes();
    char s[256];
    matMulMethod *matMul = itor->matMul;
    preCondMethod *preCond = itor->preCond;
    int sol = nodesState->rSol;
    
    if (matMul==nil) matMul = actMatMul;
    if (preCond==nil) preCond = actPreCond;

    while ((it++)<(itor->maxSteps))
      {
      	alpha = rho/sigma;
        NodeLin(cgsq,cgsu,cgsv,-alpha);
        NodeAdd(cgsh1,cgsu,cgsq);
        if (preCond->PreCond) (preCond->PreCond)(cgsh1,cgsh2); 
         else NodeAssign(cgsh2,cgsh1);
        if ((matMul->MatMul)(cgsh2,cgsv)==false) return false;
        NodeLin(cgsr,cgsr,cgsv,-alpha);
        NodeLin(sol,sol,cgsh2,alpha);

        itor->res =sqrt(NodeScalProd(cgsr,cgsr))/no;
        (itor->iteCount)++;
        if (itor->verboseP)
          {
            sprintf(s, "Solve: CGS step %d: eps=%9.3e res=%9.3e\n",
                    itor->iteCount, itor->eps, itor->res);
            ZIBStdOut(s);
          }
        if (((itor->res)<=(itor->eps))&&(it>=(itor->minSteps))) break;        

  	rhoLast = rho;

        if (rhoLast==0.0)
          {
            sprintf(s, "Solve: CGS failed rho=0.0\n");
            ZIBStdOut(s);
            return false;
          }

        rho = NodeScalProd(cgsr,cgsr0);
        beta = rho/rhoLast;
        NodeLin(cgsu,cgsr,cgsq,beta);
        NodeLin(cgsh1,cgsq,cgsp,beta);
        NodeLin(cgsp,cgsu,cgsh1,beta);
        if (preCond->PreCond) (preCond->PreCond)(cgsp,cgsh1); 
         else NodeAssign(cgsh1,cgsp);
        if ((matMul->MatMul)(cgsh1,cgsv)==false) return false;
        sigma = NodeScalProd(cgsv,cgsr0);
      }

    if (itor->verboseP)
      {
        sprintf(s, "Solve: end of CGS, n=%d, eps=%9.3e, res=%9.3e\n", 
                GetNoNodes(), itor->eps, itor->res);
        ZIBStdOut(s);
      }
    
    return true;
  }

static int CGSClose(iteMethod *itor) 
  {
    matMulMethod *matMul = itor->matMul;
    preCondMethod *preCond = itor->preCond;
    
    if (matMul==nil) matMul = actMatMul;
    if (preCond==nil) preCond = actPreCond;
    if (matMul->CloseMatMul) (matMul->CloseMatMul)();
    if (preCond->ClosePreCond) (preCond->ClosePreCond)(); 
    
    ReturnAccess(nodesState->accNode,cgsr0);
    ReturnAccess(nodesState->accNode,cgsr);
    ReturnAccess(nodesState->accNode,cgsp);
    ReturnAccess(nodesState->accNode,cgsv);
    ReturnAccess(nodesState->accNode,cgsu);
    ReturnAccess(nodesState->accNode,cgsq);
    ReturnAccess(nodesState->accNode,cgsh1);
    ReturnAccess(nodesState->accNode,cgsh2); 
    
    cgsr0 = cgsr = cgsp = cgsv = cgsu = cgsq = -1; 
    cgsh1 = cgsh2 = -1;
    
    return true;
  }

int DefCGSMethod()
  {
    return DefIteMethod("pcgs",false,false,true,true, 
                        CGSInit,CGSIterate,CGSClose);
  }
