/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "stdio.h"
#include "math.h"
 
#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "assemble.h"
#include "solve.h"

static int CholDecEnv(), CholSolEnv();
static real **ExpMatEnvelope(real**,int,int,int);
static real **ReturnMatEnvelope(real**,int);
static int *ExpVecInt(int*,int,int);
static char **ExpVecAdr(char**,int,int);
static int PrintLSEnv(printMode);
static void PrintMatEnv(char*,real**,int,int);
static int ReleaseEnv();
static int RevCutMcKee();

static real *rightSide;
static int *envIndex, dim, globStiff;
static char **nodeAdr,**saveAdr;

static int nodeCount, minDegree, rootIndex;
static char *root;

/*
   Direct solution of a positive definit matrix by a rational
   cholesky decomposition (copied from Wilkinson/ Reinsch;
   Handbook for Automatic Computation, Vol. II, Linear Algebra)
   and a forward/ backward substitution.
*/
 
/*
    CholDec(a,d,l,n) factorizes a=ldl**T
         a: symmetric positive definit Matrix
         d: vector, containing diagonal components
         l: triangular matrix, l[i][i]=1.0
         n: dimension
    error: a indefinit
 
Ref.: Wilkinson/Reinsch, p.21, routine symdet
*/
 
static int CholDecEnv()
  {
    real *a, *d=actDirector->diagonal, **l=actDirector->decomp;
    int n=actDirector->dim, *env=actDirector->env;
    char **nodeVec = actDirector->nodes, *v, **vec;
    int i, j, k, envMax, jj, lng;
    real x, y, z;
 
    Set0MatEnv(l,n,true);
    Set0VecFull(d,n);
    for (i=0; i<n; i++)
      {
        v = nodeVec[i];
        vec = NODE_NEIGHBOR_VEC(v);
        a = NODE_SPARSE_VEC(v,globStiff);
        lng = NO_OF_NODE_NEIGHBORS(v);
        jj=0;
        for  (j=env[i]; j<=i; j++)
          {
            x = (j==IV(vec[jj],iIndex))?a[jj++]:ZERO;
            if (i==j)
              {
                for (k = j-1; k>=env[i]; k--)
                  {
                    y = l[i][k-env[i]];
                    z = (l[i][k-env[i]] = y*d[k]);
                    x -= y*z;
                  }
                if (x==ZERO) return false;
                d[i] = ONE/x;
              }
            else
              {
                envMax = (env[i]>env[j])?env[i]:env[j];
                for (k = j-1; k>=envMax; k--)
                  {
                    x -= l[i][k-env[i]]*l[j][k-env[j]];
                  }
                l[i][j-env[i]]  = x;
              }
          }
      }
    for (k = 0; k<n; k++) l[k][k-env[k]] = ONE;
    return true;
  }

 /*
   CholSol(l,p,b,x,n) forward/backward substitution
     l: triangular matrix
     p: diagonal elements (d from chol_dec)
     b: rightside of ax=b
     x: solution
     n: dimension
 
Ref.: Wilkinson/Reinsch, p.22, routine symsol
*/

static int CholSolEnv()
  {
    real **l=actDirector->decomp, *p=actDirector->diagonal,
         *b=actDirector->rhs, *x=actDirector->solution;
    int n=actDirector->dim, *env=actDirector->env;
    int i, k;
    real y;
 
    Set0VecFull(x, n);
    for (i = 0; i<n; i++)      /* Forward substitution */
      {
        y = b[i];
        for (k = i-1; k>=env[i]; k--) y -= l[i][k-env[i]]*x[k];
        x[i] = y;
        b[i] = y;
      }

    for (i = n-1; i>=0; i--)   /* Backward substitution */
       {
        b[i] = b[i]*p[i];
       }

    for (i = n-1; i>=0; i--)   
      {
        x[i] = b[i];
        for (k = env[i]; k<i; k++) b[k] -= l[i][k-env[i]]*x[i];
      }

    return true;
  }

static int SetAdr(char *v)
  {
    char **vec = NODE_NEIGHBOR_VEC(v);

    envIndex[IV(v,iIndex)] = IV(vec[0],iIndex);
    rightSide[IV(v,iIndex)] =  RV(v,rRhs);

    return true;
  }

static int AssembleEnv()
  {
    real **decomp=actDirector->decomp;
    real *rhs=actDirector->rhs, *diagonal=actDirector->diagonal,
         *solution=actDirector->solution;
    int *env=actDirector->env, oldDim=actDirector->dim;
    char **nodeVec=actDirector->nodes;

    dim = GetNoNodes();

    decomp = ReturnMatEnvelope(decomp,oldDim);
	if (oldDim<dim)
      {
        actDirector->rhs = ExpVecFull(rhs, oldDim, dim);
        if ((actDirector->rhs)==nil) return false;

        actDirector->diagonal = ExpVecFull(diagonal, oldDim, dim);
        if ((actDirector->diagonal)==nil) return false;

        actDirector->solution = ExpVecFull(solution, oldDim, dim);
        if ((actDirector->solution)==nil) return false;

        actDirector->env = ExpVecInt(env, oldDim, dim); /* initializes env[k]=k */
        if ((actDirector->env)==nil) return false;

        actDirector->nodes = ExpVecAdr(nodeVec, oldDim, dim); /* initializes env[k]=k */
        if ((actDirector->nodes)==nil) return false;

        actDirector->dim = dim;
      }
	if (actMatMul->holdsSparseMatrix)
	  globStiff = (actMatMul->InitMatMul)(-1,actProblem->LocAss,
	                         nodesState->rRhs,actProblem->LocAssB);
	else
	  globStiff = InitSparseMul(-1,actProblem->LocAss,
	                         nodesState->rRhs,actProblem->LocAssB);
    actDirector->stiffSparse = globStiff;
    dim = actDirector->dim;
    rightSide = actDirector->rhs;
    envIndex = actDirector->env;
    nodeAdr = actDirector->nodes;
    RevCutMcKee();
    ApplyNode(SetAdr,all);
    actDirector->decomp = ExpMatEnvelope(decomp, oldDim, dim, theProblem->symP);
    if ((actDirector->decomp)==nil) return false;

    return true;
  }

int InitEnv(int stiff)
  {
    real **decomp=actDirector->decomp;
    real *rhs=actDirector->rhs, *diagonal=actDirector->diagonal,
         *solution=actDirector->solution;
    int *env=actDirector->env, oldDim=actDirector->dim;
    char **nodeVec=actDirector->nodes;

    dim = GetNoNodes();

    if (oldDim<dim)
      {
        actDirector->rhs = ExpVecFull(rhs, oldDim, dim);
        if ((actDirector->rhs)==nil) return false;

        actDirector->diagonal = ExpVecFull(diagonal, oldDim, dim);
        if ((actDirector->diagonal)==nil) return false;

        actDirector->solution = ExpVecFull(solution, oldDim, dim);
        if ((actDirector->solution)==nil) return false;

        actDirector->env = ExpVecInt(env, oldDim, dim); /* initializes env[k]=k */
        if ((actDirector->env)==nil) return false;

        actDirector->nodes = ExpVecAdr(nodeVec, oldDim, dim); 
        if ((actDirector->nodes)==nil) return false;

        actDirector->dim = dim;
      }
    globStiff = stiff;

    dim = actDirector->dim;
    rightSide = actDirector->rhs;
    envIndex = actDirector->env;
    nodeAdr = actDirector->nodes;
    RevCutMcKee();
    ApplyNode(SetAdr,all);
    actDirector->decomp = ExpMatEnvelope(decomp, oldDim, dim, theProblem->symP);
    if ((actDirector->decomp)==nil) return false;

    return true;
  }

static int PrintLSEnv(printMode mode)
  {
    real **decomp=actDirector->decomp;
    real *rhs=actDirector->rhs, *diagonal=actDirector->diagonal,
         *solution=actDirector->solution;
    int dim=actDirector->dim;

    switch (mode)
    {
    case printLinSys:
      PrintMatSparse("Stiffness matrix", actDirector->stiffSparse,
                     dim, SYMMETRIC);
      PrintVecFull("Right side",rhs,dim);
      break;
    case printDecomp:
      PrintMatEnv("L matrix", decomp, dim, SYMMETRIC);
      PrintVecFull("Diagonal",diagonal,dim);
      break;
    case printSol:
      PrintVecFull("Solution",solution,dim);
      break;
    case printNothing:
      break;
    }

    return true;
  }

static int ReleaseEnv()
  {
    int oldDim = actDirector->dim;

	rightSide = actDirector->rhs = ExpVecFull(actDirector->rhs,oldDim,-1);
    rightSide = actDirector->diagonal = ExpVecFull(actDirector->diagonal,oldDim,-1);
    actDirector->solution = ExpVecFull(actDirector->solution,oldDim,-1);
    actDirector->decomp = ReturnMatEnvelope(actDirector->decomp,oldDim);
    envIndex = actDirector->env = ExpVecInt(actDirector->env, oldDim, -1);
    nodeAdr = actDirector->nodes = ExpVecAdr(actDirector->nodes, oldDim, -1); 
    actDirector->dim = 0;

	return true;
  }

int DefEnvChol()
  {
    return DefDirMethod("envchol","Cuthill-McKee",
                        envelope,true,true,CholDecEnv,CholSolEnv,
                        (int(*)())nil,
                        AssembleEnv,(int(*)())nil,PrintLSEnv,ReleaseEnv);
  }

static real **ExpMatEnvelope(real **x,int oldLng,int newLng,int symP)
  {
    int k, kenv;
    real **y=nil;

	if (newLng>0)
	  {
		y = (real**)ZIBAlloc((long)newLng*sizeof(real*));
		if (y==nil)
		  { 
			ZIBStdOut("Assemble: not enough memory (ExpMatEnvelope)\n"); 
			return nil; 
		  }
	
		for(k = 0; k<newLng; k++)
		  {
			kenv = k - envIndex[k] + 1;
			y[k]=(real*)GetSmallElem(kenv*sizeof(real));
			if (y[k]==nil)
			  { 
				ZIBStdOut("Assemble: not enough memory (ExpMatEnvelope)\n"); 
				return nil; 
			  }
		  }
      }
	return y;
  }

static real **ReturnMatEnvelope(real **x,int lng)
  {
    int k, kenv;

	if (x==nil) return nil;
	for(k = 0; k<lng; k++)
	  {
		kenv = k - envIndex[k] + 1;
		ReturnSmallElem(kenv*sizeof(real),(ptr*)x[k]);
	  }
	ZIBFree((ptr)x);

	return nil;
  }
 
static int *ExpVecInt(int *x,int oldLng,int newLng)
  {
    int k;
    int *y = nil;
 
	if (newLng>0)
	  {
		y = (int*)ZIBAlloc((long)newLng*sizeof(int));
		if (y==nil)
		  { ZIBStdOut("Assemble: not enough memory (ExpVecInt)\n"); return nil; }
		for(k = 0; k<newLng; k++) y[k] = k;
	  }
    if (x!=nil) ZIBFree((ptr)x);
    return y;
  }
 
static char **ExpVecAdr(char **x,int oldLng,int newLng)
  {
    int k;
    char **y = nil;
 
	if (newLng>0)
	  {
		y = (char**)ZIBAlloc((long)newLng*sizeof(char*));
		if (y==nil)
		  { ZIBStdOut("Assemble: not enough memory (ExpVecAdr)\n"); return nil; }
		for(k = 0; k<newLng; k++) y[k] = nil;
      }
    if (x!=nil) ZIBFree((ptr)x);
	return y;
  }

void Set0MatEnv(real **x, int n, int symp)
  {
    int i,k;
 
    for(i = 0; i<n; i++)
      for(k = 0; k<=i-envIndex[i]; k++) x[i][k] = ZERO;
    return;
  }
 
static void PrintMatEnv(char *s, real **a, int n, int triP)
  {
    int i,k;
 
    sprintf(globBuf,"Envelope matrix '%s(%s)', dim=%d\n", s,
            (triP==SYMMETRIC)?"SYMMETRIC":"FULL", n);
    ZIBStdOut(globBuf);
    for(i = 0; i<n; i++)
      for(k = envIndex[i]; k<=((triP==SYMMETRIC)?i:n-1); k++)
        if (a[i][k-envIndex[i]]!=ZERO)
          {
            sprintf(globBuf,"  (% 3d,% 3d):% 8.3e\n", i, k, a[i][k-envIndex[i]]);
            ZIBStdOut(globBuf);
          }
    return;
  }

static int PreCuthillMcKee(char *v)
  {
    int lng = NO_OF_NODE_NEIGHBORS(v), index = IV(v,iIndex);

    nodeAdr[nodeCount++] = v;
    IV(v,iIndex) = -1;
    if (CV(v,cBoundP)!=INTERIOR)
      if (lng<=minDegree)
        { minDegree = lng; root = v; rootIndex = index; }
    return true;
  }

static int PostCuthillMcKee(char *v)
  {
    int i, k, lng = NO_OF_NODE_NEIGHBORS(v);
    char **vec = NODE_NEIGHBOR_VEC(v), *c;
    real *stiff = NODE_SPARSE_VEC(v, globStiff), x;

    for (i=0; i<lng-1; i++)
      for (k=i+1; k<lng; k++)
        if (IV(vec[i],iIndex)>IV(vec[k],iIndex))
          {
            c = vec[i]; vec[i] = vec[k]; vec[k] = c;
            x = stiff[i]; stiff[i] = stiff[k]; stiff[k] = x;
          }

    return true;
  }

static int NumberNeighbors(char *v)
  {
    int k, lng = NO_OF_NODE_NEIGHBORS(v);
    char **vec = NODE_NEIGHBOR_VEC(v);

    for (k=0; k<lng; k++)
      if (IV(vec[k],iIndex)<0)
        {
          saveAdr[nodeCount] = vec[k];
          IV(vec[k],iIndex) = nodeCount--;
        }

    return true;
  }

static int RevCutMcKee()
  {
    int noOfNodes, k;

    saveAdr=nodeAdr;
    noOfNodes = GetNoNodes();
    nodeAdr = (char**)ZIBAlloc(noOfNodes*sizeof(char*));
    if (nodeAdr==nil) return false;
    nodeCount = 0; minDegree = 10000; root = nil;
    ApplyNode(PreCuthillMcKee,all);

    nodeCount = noOfNodes-1;
    IV(root,iIndex) = nodeCount;
    saveAdr[nodeCount] = root;
    nodeCount--;
    for (k=noOfNodes-1; k>=0; k--)
      {
        NumberNeighbors(saveAdr[k]);
      }
    if (nodeCount!=-1)
      { ZIBStdOut("Something bad happened\n"); return false; }
    ZIBFree(nodeAdr);
    nodeAdr = saveAdr;

    for (k=0; k<noOfNodes; k++) PostCuthillMcKee(nodeAdr[k]);

    return true;
  }
