/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "stdio.h"
#include "math.h"
 
#include "zibutil.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "assemble.h"
#include "solve.h"

static int AssFull(TR*),SetBCFull(char*);
static int CholDec(), CholSol(), ChkSol();

static real **assemb, *rightSide;
static int dim;

/*
   Direct solution of a positive definit matrix by a rational
   cholesky decomposition (copied from Wilkinson/ Reinsch;
   Handbook for Automatic Computation, Vol. II, Linear Algebra)
   and a forward/ backward substitution.
*/
 
/*
    CholDec(a,d,l,n) factorizes a=ldl**T
         a: symmetric positive definit Matrix
         d: vector, containing diagonal components
         l: triangular matrix, l[i][i]=1.0
         n: dimension
    error: a indefinit
 
Ref.: Wilkinson/Reinsch, p.21, routine symdet
*/
 
static int CholDec()
  {
    REAL **a=actDirector->stiff, *d=actDirector->diagonal, **l=actDirector->decomp;
    int n=actDirector->dim;
    int i, j, k;
    REAL x, y, z;
 
    Set0MatFull(l, n, true);
    Set0VecFull(d, n);
    for (i = 0; i<n; i++)
      for (j = 0; j<i+1; j++)
        {
          x = a[i][j];
          if (i==j)
            {
              for (k = j-1; k>=0; k--)
                {
                  y = l[i][k];
                  z = (l[i][k] = y*d[k]);
                  x -= y*z;
/* if (z!=ZERO)
{
sprintf(globBuf,"z=l[%d][%d]=%e, x=%e\n",i,k,z,x);
ZIBStdOut(globBuf);
} */
                }
              if (x==ZERO) return false;
              d[i] = ONE/x;
/* sprintf(globBuf,"d[%d]=%e\n",i,d[i]);
ZIBStdOut(globBuf); */
            }
          else
            {
              for (k = j-1; k>=0; k--)
                {
                  x -= l[i][k]*l[j][k];
/* if ((l[i][k]*l[j][k])!=ZERO)
{
sprintf(globBuf,"l[%d][%d]=%e, l[%d][%d]=%e, x=%e\n",i,k,l[i][k],j,k,l[j][k],x);
ZIBStdOut(globBuf);
} */
                }
/* if (x!=ZERO) { sprintf(globBuf,"l[%d][%d]=%e\n",i,j,x);
ZIBStdOut(globBuf); } */
              l[i][j]  = x;
            }
        }
    for (k = 0; k<n; k++) l[k][k] = ONE;
    return true;
  }

 /*
   CholSol(l,p,b,x,n) forward/backward substitution
     l: triangular matrix
     p: diagonal elements (d from chol_dec)
     b: rightside of ax=b
     x: solution
     n: dimension
 
Ref.: Wilkinson/Reinsch, p.22, routine symsol
*/
 
static int CholSol()
  {
    REAL **l=actDirector->decomp, *p=actDirector->diagonal, *b=actDirector->rhs,
         *x=actDirector->solution;
    int n = actDirector->dim;
    int i, k;
    REAL y;
 
    Set0VecFull(x, n);
    for (i = 0; i<n; i++)      /* Forward substitution */
      {
        y = b[i];
        for (k = i-1; k>=0; k--) y -= l[i][k]*x[k];
        x[i] = y;
      }
    for (i = n-1; i>=0; i--)   /* Backward substitution */
      {
        y = x[i]*p[i];
        for (k = i+1; k<n; k++) y -= l[k][i]*x[k];
        x[i] = y;
      }
    return true;
  }
 
 static int ChkSol()
  {
    REAL **a = actDirector->stiff, *b = actDirector->rhs, *x = actDirector->solution,
         *d = actDirector->diagonal;
    int n = actDirector->dim, verboseP = actDirector->verboseP;
    int i, k;
    REAL s, amax;
    REAL zmin = RMAX, zmax = RMIN;

    amax = ZERO;
    for (i = 0; i<n; i++)
      {
        s = ZERO;
        for (k = 0; k<n; k++) s += ((i<k)?a[k][i]:a[i][k])*x[k];
        if (fabs(s-b[i])>amax) amax = fabs(s-b[i]);
      }
/*
    sprintf(globBuf, "Solve: maximal deviation %e (ChkSol)\n",
            amax);
    ZIBStdOut(globBuf);
*/
    for (k = 0; k<n; k++)
      {
        if (d[k]<zmin) zmin = d[k];
        if (d[k]>zmax) zmax = d[k];
      }
    if (verboseP)
      {
        sprintf(globBuf,"Solve: min/max(diag) %9.3e/%9.3e (ChkSol)\n",
                zmin, zmax);
        ZIBStdOut(globBuf);
      }
    return zmin>=ZERO;
  }

int AssembleFull()
  {
    real **stiff = actDirector->stiff, **decomp = actDirector->decomp;
    real *rhs = actDirector->rhs, *diagonal = actDirector->diagonal,
         *solution = actDirector->solution;
    int oldDim = actDirector->dim;

    dim = GetNoNodes();
    if (oldDim!=dim)
      {
        actDirector->stiff = ExpMatFull(stiff, oldDim, dim, theProblem->symP);
        if ((actDirector->stiff)==nil) return false;

        actDirector->rhs = ExpVecFull(rhs, oldDim, dim);
        if ((actDirector->rhs)==nil) return false;

        actDirector->decomp = ExpMatFull(decomp, oldDim, dim, theProblem->symP);
        if ((actDirector->decomp)==nil) return false;

        actDirector->diagonal = ExpVecFull(diagonal, oldDim, dim);
        if ((actDirector->diagonal)==nil) return false;

        actDirector->solution = ExpVecFull(solution, oldDim, dim);
        if ((actDirector->solution)==nil) return false;

        actDirector->dim = dim;
      }
    Set0MatFull(actDirector->stiff, dim, theProblem->symP);
    Set0VecFull(actDirector->rhs, dim);

    dim = actDirector->dim;
    assemb = actDirector->stiff;
    rightSide = actDirector->rhs;

    if (!ApplyT(AssFull, all)) return false;
    if (!(nodesState->compNormP))
      if (!ApplyNode(SetBCFull, dirichlet)) return false;

    return true;
  }

int PrintLSFull(printMode mode)
  {
    real **stiff = actDirector->stiff, **decomp = actDirector->decomp;
    real *rhs = actDirector->rhs, *diagonal = actDirector->diagonal,
         *solution = actDirector->solution;
    int dim = actDirector->dim;

    switch (mode)
    {
    case printLinSys:
      PrintMatFull("Stiffness matrix", stiff, dim, SYMMETRIC);
      PrintVecFull("Right side",rhs,dim);
      break;
    case printDecomp:
      PrintMatFull("L matrix", decomp, dim, SYMMETRIC);
      PrintVecFull("Diagonal",diagonal,dim);
      break;
    case printSol:
      PrintVecFull("Solution",solution,dim);
      break;
    case printNothing:
      break;
    }

    return true;
  }

int ReleaseFull()
  {
    actDirector->stiff = ExpMatFull(actDirector->stiff,actDirector->dim,-1,SYMMETRIC);
    actDirector->decomp = ExpMatFull(actDirector->decomp,actDirector->dim,-1,SYMMETRIC);
    return true;
  }

int DefFullChol()
  {
    return DefDirMethod("fullchol","norenumb",
                        full,true,true,CholDec,CholSol,ChkSol,
                        AssembleFull,(int(*)())nil,PrintLSFull,ReleaseFull);
  }

static int AssFull(TR *t)
  {
    int i,k,index[100],i1,i2,no;
    real **assA;
    real *assB;
    char *vs[100];
 
    no = GetNodeAddresses(t, vs, 100);
    if (no<=0) return false;
    for (i=0; i<no; i++) index[i] = IV(vs[i],iIndex); 

    assA = (actProblem->LocAss)(t);
    assB = (actProblem->LocAssB)(t);
    if ((assA==nil)||(assB==nil)) return false;
    for(i = 0; i<no; i++)
      {
        i1  =index[i];
        for(k = 0; k<=((theProblem->symP==SYMMETRIC)?i:no-1); k++)
          {
            i2 = index[k];
            if (theProblem->symP==SYMMETRIC)
              {
                if (i1<i2) assemb[i2][i1] += assA[i][k];
                else       assemb[i1][i2] += assA[i][k];
              }
            else assemb[i2][i1] += assA[k][i];
          }
        rightSide[i1] += assB[i];
      }
    return true;
  }

static int SetBCFull(char *v)
  {
    int i = IV(v,iIndex),k;
    real bxy;

    (actProblem->Dirichlet)(RV(v,rX),RV(v,rY),CV(v,cClassA),&bxy,CV(v,cEqnNo));
    for (k = 0; k<dim; k++)
      {
        if (theProblem->symP==SYMMETRIC)
          {
            if(i>k)
              {
                rightSide[k] -= bxy*assemb[i][k];
                assemb[i][k] = ZERO;
              }
            if(i<k)
              {
               rightSide[k] -= bxy*assemb[k][i];
               assemb[k][i] = ZERO;
              }
           }
         else
           {
            if(i!=k)
              {
                rightSide[k] -= bxy*assemb[k][i];
                assemb[i][k] = ZERO;
                assemb[k][i] = ZERO;
              }
           }
      }
    rightSide[i] = bxy;
    assemb[i][i] = ONE;
    return true;
  }

void Set0MatFull(real **x, int n, int symp)
  {
    int i,k;
 
    for(i = 0; i<n; i++)
      for(k = 0; k<((symp==SYMMETRIC)?i+1:n); k++) x[i][k] = ZERO;
    return;
  }

void Set0VecFull(real *x,int n)
  {
    int k;
 
    for(k = 0; k<n; k++) x[k] = ZERO;
    return;
  }
