/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "assemble.h"
#include "access.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"

static real *hbg = nil, *cGivens = nil, *sGivens = nil;
static real **R = nil;
static int gmres = -1;
static int *v = nil;
static int ConvTest(real,real), GMResGivens(real*,int), GMResRSolve(int);

static int GMResInit(iteMethod *itor) 
  {
    char s[256];
    matMulMethod *matMul = itor->matMul;
    preCondMethod *preCond = itor->preCond;
    int i, maxOrtho = itor->maxOrtho;

    if (matMul==nil) matMul = actMatMul;
    if (preCond==nil) preCond = actPreCond;

    itor->verboseP = true;

    if (gmres==-1)
     {
      gmres = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"gmres");
      if (gmres==-1)
        {
          ZIBStdOut("GMResInit: not enough memory for gmres\n");
          return false;
        }
      NodeSetZero(gmres);
     }
    if (hbg==nil)
      hbg = ExpVecFull((real*)nil,0,maxOrtho+1);
    if (cGivens==nil)
     {
      cGivens = ExpVecFull((real*)nil,0,maxOrtho);
      for (i=0;i<maxOrtho;i++) cGivens[i] = 0.0;
     }
    if (sGivens==nil)
     {
      sGivens = ExpVecFull((real*)nil,0,maxOrtho);
      for (i=0;i<maxOrtho;i++) sGivens[i] = 0.0;
     }
    if (R==nil)
      R = ExpMatFull((real**)nil,0,maxOrtho,FULL);

    if (v==nil)
     {
      v = (int*)ZIBAlloc((maxOrtho+1)*sizeof(int));
      if (v==nil) return false;
      for (i=0;i<maxOrtho+1;i++) 
        {
          v[i] = GetAccess(nodesState->accNode,sizeof(real),sizeof(real),"gmres_v");
          if (v[i]==-1)
           {
             ZIBStdOut("GMResInit: not enough memory for gmres_v\n");
             return false;
           }
        }
     }

    itor->maxRestarts = (itor->maxSteps)/(itor->maxOrtho) + 1;

    if (itor->verboseP)
      {
        sprintf(s, "Solve: start of GMRes(%d), n=%d, eps=%9.3e, steps=(%d,%d), maxRestarts=%d, %s\n", 
                maxOrtho, GetNoNodes(), itor->eps, itor->minSteps, itor->maxSteps, itor->maxRestarts,
                preCond->name);
        ZIBStdOut(s);
      }
    
    if (matMul->InitMatMul)
      (matMul->InitMatMul)(-1,actProblem->LocAss,nodesState->rRhs, 
                           actProblem->LocAssB);
    if (preCond->InitPreCond) (preCond->InitPreCond)(); 
        
    itor->iteCount = 0;
    
    return true;
  }

static int GMResIterate(iteMethod *itor) 
  {
    int restarts = -1, i, k, kMax = 0;
    int no = GetNoNodes(), maxOrtho = itor->maxOrtho;
    int rSol = nodesState->rSol, rRhs = nodesState->rRhs;
    real tmp, h, r;
    real tol = itor->eps;
    char s[256], done = false;
    matMulMethod *matMul = itor->matMul;
    preCondMethod *preCond = itor->preCond;
    
    if (matMul==nil) matMul = actMatMul;
    if (preCond==nil) preCond = actPreCond;

/*               outer GMRes iteration    
*/
    while ((++restarts)<(itor->maxRestarts))
      {
      	if ((matMul->MatMul)(rSol,gmres)==false) return false;
        NodeSub(gmres,rRhs,gmres);
        if (preCond->PreCond) (preCond->PreCond)(gmres,v[0]);
         else NodeAssign(v[0],gmres); 
        for (i=0;i<maxOrtho+1;i++) hbg[i] = 0.0;
        hbg[0] = sqrt(NodeScalProd(v[0],v[0]));
        if (ConvTest(fabs(hbg[0]),tol)==true) break;
        NodeFacAssign(v[0],v[0],1.0/hbg[0]);

/*               inner GMRes-iteration
*/
        for (k=1;k<maxOrtho+1;k++)
          {   
            kMax = k;     
            if ((matMul->MatMul)(v[k-1],gmres)==false) return false;
            if (preCond->PreCond) (preCond->PreCond)(gmres,v[k]);
             else NodeAssign(v[k],gmres);

/*               Gram-Schmidt process
*/
            for (i=0;i<k;i++) R[k-1][i] = NodeScalProd(v[i],v[k]);
            for (i=0;i<k;i++) NodeLin(v[k],v[k],v[i],-R[k-1][i]);
            h = sqrt(NodeScalProd(v[k],v[k]));
            NodeFacAssign(v[k],v[k],1.0/h);

/*               Apply previous Givens rotations
*/
            for (i=0;i<k-1;i++) GMResGivens(R[k-1],i);

/*               Compute new Givens coefficients 
*/
            r = R[k-1][k-1];
            tmp = sqrt(r*r+h*h);
            cGivens[k-1] = r/tmp;
            sGivens[k-1] = -h/tmp;
            R[k-1][k-1] = tmp;

/*               Update vector hbg with new Givens coefficients
*/
            GMResGivens(hbg,k-1);

            (itor->iteCount)++;
            if (ConvTest(fabs(hbg[k]),tol)==true)
              {
                done = true; break;
              }
          }

/*    	         Compute solution "R^(-1) times hbg" of least squares problem
*/
  	GMResRSolve(kMax);

/*               Update solution vector 
*/
        for (i=0;i<kMax;i++) NodeLin(rSol,rSol,v[i],hbg[i]);

        itor->res = fabs(hbg[kMax])/no;

/*               Print informations
*/
        if ((itor->verboseP)&&(!done))
         {
	   sprintf(s,"Solve: GMRes(%d) %d. restart: res=%9.3e\n",
                      itor->maxOrtho,restarts+1,itor->res);
           ZIBStdOut(s);
         } 

        if (((done)||(restarts==(itor->maxRestarts)))&&(itor->verboseP))
         {
           sprintf(s, "Solve: end of GMRes(%d): steps=%d eps=%9.3e res=%9.3e\n", 
                maxOrtho, itor->iteCount, itor->eps, itor->res);
           ZIBStdOut(s);
         }

        if (done) break;
      } 
    
    return true;
  }

static int GMResClose(iteMethod *itor) 
  {
    matMulMethod *matMul = itor->matMul;
    preCondMethod *preCond = itor->preCond;
    int maxOrtho = itor->maxOrtho, i;
    
    if (matMul==nil) matMul = actMatMul;
    if (preCond==nil) preCond = actPreCond;
    if (matMul->CloseMatMul) (matMul->CloseMatMul)();
    if (preCond->ClosePreCond) (preCond->ClosePreCond)(); 
    
    ReturnAccess(nodesState->accNode,gmres);
    ZIBFree(hbg);
    ZIBFree(cGivens);
    ZIBFree(sGivens);
    ExpMatFull(R,maxOrtho,0,FULL);
    for (i=0;i<maxOrtho+1;i++) 
      ReturnAccess(nodesState->accNode,v[i]);
    ZIBFree(v);

    gmres = -1; 
    hbg = nil;
    cGivens = nil;
    sGivens = nil;
    R = nil; 
    v = nil;
 
    return true;
  }

static int ConvTest(real g, real tol)
  {
    char s[256];
    int no = GetNoNodes();

    if (actItor->verboseP)
     {
	sprintf(s,"Solve: GMRes(%d) step %d: norm(g)=%9.3e\n",
                   actItor->maxOrtho,actItor->iteCount,g/no);
        ZIBStdOut(s);
     }
    if (g/no<tol) return true;
    else return false;
  }

static int GMResGivens(real *vec, int row)
  {
    real tmp;

    tmp = cGivens[row]*vec[row] - sGivens[row]*vec[row+1];
    vec[row+1] = sGivens[row]*vec[row] + cGivens[row]*vec[row+1];
    vec[row] = tmp;

    return true;
  }

static int GMResRSolve(int k)
  {
    int i, j;
    
    for (j=k-1;j>=0;j--)
     {
       hbg[j] = hbg[j];
       for (i=j+1;i<k;i++) hbg[j] -= R[i][j]*hbg[i];
       hbg[j] /= R[j][j];
     }

    return true;
  }

int DefGMResMethod()
  {
    return DefIteMethod("pgmres",false,false,true,true, 
                        GMResInit,GMResIterate,GMResClose);
  }
