/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include "zibutil.h"
#include "fixedsize.h"
#include "access.h"
#include "appevents.h"
#include "triang2.h"
#include "nodes.h"

/*
 * This file contains routines for a nested dissection and a reverse
 * Cuthill-McKee ordering of the nodal points. The algorithms are implemented
 * using as much of the given data structure as possible. Thus the
 * implementation is not tuned to be optimal.
 */

static int noOfPtNeigh = -1, adrOfPtNeigh = -1;
static sortMode globSortMode = sortIndex;

static int      levelNo, treeSize, compLevel, minDeg, lastNo, more;
static int      treeDepth, levelAt; 
static PT       *minPoint;
static PT       **pointOfNumber;
static PT       ***subGraph;
static int      structNextNo, structLastNo, subGraphNo[100];
static PT       **levelStructNext, **levelStructLast, **levelStruct;

static int treeDepthPt = -1, levelPt = -1;

static int RenumberPts(), RetPtNeighbor();
static int InitPtNeigh(PT*), AllocPtVec(PT*), RetPtNeigh(PT*), ClearLevel(PT*),
           SetTreeDepthZero(PT*), Mark(PT*), ResetMarkOnLS(PT*), CompDeg(PT*),
           RootLS(PT*), RootLsRCM(PT*), Seperate(PT*), NumberTheNeighbors(PT*);
static int CountPtNeighb(EDG*), StorePtNeigh(EDG*);
static int GenPtNeighbor(sortMode);
static int ApplyToNeighbors(int (*)(PT*,int),int,PT*,int);
static int SetLevel(PT*,int), SetLevelRCM(PT*,int), TestNeighbors(PT*,int);
static PT* MinDegOnLevel(int), *MinDegOnLevelRCM(int);
static PT* FindRootRCM(PT*), *FindRootND(PT*);
static void ChangeIndex(PT*,PT*);
static void NumberThePoint(PT*);

static int GenPtNeighbor(sortMode mode)
  {
    int newP = (noOfPtNeigh<0);

    if (noOfPtNeigh<0)
      noOfPtNeigh = GetAccess(nodesState->accPoint,sizeof(int),
                              sizeof(int),"# of neighbor points");
    if (noOfPtNeigh<0) return false;
    if (adrOfPtNeigh<0)
      adrOfPtNeigh = GetAccess(nodesState->accPoint,sizeof(PT*),
                               sizeof(PT*),"adr of neighbor points");
    if (adrOfPtNeigh<0)
      {
        ReturnAccess(nodesState->accPoint,noOfPtNeigh);
        noOfPtNeigh = -1;
        return false;
      }
    globSortMode = mode;
    if (newP) ApplyP(InitPtNeigh,all);
    else ApplyP(RetPtNeigh,all);
    ApplyE(CountPtNeighb,all);
    if (!ApplyP(AllocPtVec,all)) return false;
    ApplyE(StorePtNeigh,all);

    return true;
  }

static int RetPtNeighbor()
  {
    ApplyP(RetPtNeigh,all);
    ReturnAccess(nodesState->accPoint,noOfPtNeigh);
    noOfPtNeigh = -1;
    ReturnAccess(nodesState->accPoint,adrOfPtNeigh);
    adrOfPtNeigh = -1;
    return true;
  }

static int InitPtNeigh(PT *p)
  {
    IP(p,noOfPtNeigh) = 0;
    AP(p,adrOfPtNeigh) = nil;
    return true;
  }

static int RetPtNeigh(PT *p)
  {
    int lng=NO_OF_PT_NEIGHBORS(p);
    PT **vec=PT_NEIGHBOR_VEC(p);

    if (vec) ReturnSmallElem(lng*sizeof(PT*),(ptr*)vec);
    IP(p,noOfPtNeigh) = 0;
    AP(p,adrOfPtNeigh) = nil;

    return true;
  }

static int CountPtNeighb(EDG* ed)
  {
    PT *p1=ed->p1, *p2=ed->p2;

    IP(p1,noOfPtNeigh)++; IP(p2,noOfPtNeigh)++;
    return true;
  }

static int AllocPtVec(PT *p)
  {
    int lng;
    PT** vec;
    
    IP(p,noOfPtNeigh)++;
    lng=IP(p,noOfPtNeigh);

    vec = (PT**)GetSmallElem(lng*sizeof(PT*));
    if (vec==nil) return false;
    AP(p,adrOfPtNeigh) = (ptr)vec;
    vec[0] = p;
    return true;
  }

static int StorePtNeigh(EDG* ed)
  {
    PT *p1=ed->p1, *p2=ed->p2, *p=nil;
    int lng1=NO_OF_PT_NEIGHBORS(p1), lng2=NO_OF_PT_NEIGHBORS(p2);
    PT **vec1=PT_NEIGHBOR_VEC(p1), **vec2=PT_NEIGHBOR_VEC(p2);
    int i, k, pos=0, boolP, lng;

/* Find insertion point */
    for (k=0; k<lng1; k++)
      {
        p = vec1[k];
        if (p==nil) { pos = k; break; }
        lng = NO_OF_PT_NEIGHBORS(p);
        boolP = (globSortMode==sortDepth)?
                  (lng<lng2):((p->indexP)>(p2->indexP));
        if (boolP) { pos = k; break; }
      }
/* Find number of stored neighbors */
    if (p!=nil)
      for (; k<lng1; k++)
        if (vec1[k]==nil) break;
    if (k==lng1)
      { ZIBStdOut("Wrong NO_OF_PT_NEIGHBORS\n"); return false; }
/* Move upper neighbors */
    for (i=k; i>pos; i--) vec1[i] = vec1[i-1];
/* Insert new neigbor */
    vec1[pos] = p2;

    for (k=0; k<lng2; k++)
      {
        p = vec2[k];
        if (p==nil) { pos = k; break; }
        lng = NO_OF_PT_NEIGHBORS(p);
        boolP = (globSortMode==sortDepth)?
                  (lng<lng1):((p->indexP)>(p1->indexP));
        if (boolP) { pos = k; break; }
      }
    if (p!=nil)
      for (; k<lng2; k++)
        if (vec2[k]==nil) break;
    if (k==lng2)
      { ZIBStdOut("Wrong NO_OF_PT_NEIGHBORS\n"); return false; }
    for (i=k; i>pos; i--) vec2[i] = vec2[i-1];
    vec2[pos] = p1;

    return true;
  }

static int ApplyToNeighbors(int (*Routine)(PT*, int),int to,PT *of,int data)
  {
    PT *p, **vec;
    int i, rc = true, noOfNeigh = NO_OF_PT_NEIGHBORS(of);

    switch (to) {
    case all:
       vec = PT_NEIGHBOR_VEC(of);
       for (i = 0; i < noOfNeigh; i++)
         {
           p = vec[i];
           rc = (*Routine) (p, data);
           if (!rc)
              break;
         }
        break;
    }

   return rc;
  }

static int SetLevel(PT *p, int level)
  {
    int locTreeDepth = IP(p,treeDepthPt);

    if ((p->mark) && (locTreeDepth == treeDepth))
     {
        p->mark = false;
        IP(p,levelPt) = level;
        treeSize++;
        more = true;
        structNextNo++;
        levelStructNext[structNextNo] = p;
     }
    return true;
  }

static int SetLevelRCM(PT *p, int level)
  {
    if ( p->mark )
     {
        p->mark = false;
        IP(p,levelPt) = level;
        treeSize++;
        more = true;
        structNextNo++;
        levelStructNext[structNextNo] = p;
     }
    return true;
}

static int ClearLevel(PT *p)
  {
    IP(p,levelPt) = 0;

    return true;
  }

static int SetTreeDepthZero(PT *p)
  {
    IP(p,treeDepthPt) = 0;

    return true;
  }

static int Mark(PT *p)
  {
    p->mark = true;

    return true;
  }

static int ResetMarkOnLS(PT *p)
  {
    int lev = IP(p,levelPt);

    if ( lev > 0  )
        p->mark = true;

    return true;
  }

static int RootLS(PT *root)
  {
    PT *p;
    int i;

    levelNo = 0;
    treeSize = 0;

/*    ApplyP(ClearLevel, all);   */
    for (i=0; i<subGraphNo[treeDepth]; i++)
       {
         p = subGraph[treeDepth][i];
         IP(p,levelPt) = 0;
       }

    SetLevel(root, levelNo + 1);
    while (more)
     {
        levelNo++;
        more = false;
        levelStruct     = levelStructNext;
        levelStructNext = levelStructLast;
        levelStructLast = levelStruct;
        structLastNo    = structNextNo;
        structNextNo    = -1;

        for (i=0; i<=structLastNo; i++)
          {
            p = levelStructLast[i];
            ApplyToNeighbors(SetLevel, all, p, levelNo + 1);
          }
     }

/*    ApplyP(ResetMarkOnLS, all);   */
    for (i=0; i< subGraphNo[treeDepth]; i++)
       {
         p = subGraph[treeDepth][i];
         ResetMarkOnLS(p);
       }

    return levelNo;
  }

static int RootLsRCM(PT *root)
  {
    PT *p;
    int i;

    levelNo = 0;
    treeSize = 0;

    ApplyP(ClearLevel, all);   

    SetLevelRCM(root, levelNo + 1);
    while (more)
     {
        levelNo++;
        more = false;
        levelStruct     = levelStructNext;
        levelStructNext = levelStructLast;
        levelStructLast = levelStruct;
        structLastNo    = structNextNo;
        structNextNo    = -1;

        for (i=0; i<=structLastNo; i++)
          {
            p = levelStructLast[i];
            ApplyToNeighbors(SetLevelRCM, all, p, levelNo + 1);
          }
     }

    ApplyP(ResetMarkOnLS, all);

    return levelNo;
  }

static int CompDeg(PT *p)
  {
    int lev = IP(p,levelPt), noOfNeigh = NO_OF_PT_NEIGHBORS(p);

    if ( lev != compLevel ) return true;

    if (noOfNeigh <= minDeg)
      {
        minDeg = noOfNeigh;
        minPoint = p;
      }
    return true;
  }

static PT* MinDegOnLevel(int level)
  {
    int i;
    PT  *p;

    minDeg = actTriang->noOfPoints;
    compLevel = level;
    minPoint = nil;

/*   ApplyP(CompDeg, all);   */
    for (i=0; i<subGraphNo[treeDepth]; i++)
       {
         p = subGraph[treeDepth][i];
         CompDeg(p);
       }

    return minPoint;
  }

static PT* MinDegOnLevelRCM(int level)
  {
    minDeg = actTriang->noOfPoints;
    compLevel = level;
    minPoint = nil;

    ApplyP(CompDeg, all);   

    return minPoint;
  }
  
static PT* FindRootRCM(PT *root)
  {
    int noOfLevels, newNoOfLevels;

    newNoOfLevels = RootLsRCM(root);
    if (treeSize == newNoOfLevels)
       return root;
    do {
        noOfLevels = newNoOfLevels;
        root = MinDegOnLevelRCM(noOfLevels);
        newNoOfLevels = RootLsRCM(root);

        if (treeSize == newNoOfLevels)
           return root;

     } while (newNoOfLevels > noOfLevels);
 
    return root;
  }

static PT* FindRootND(PT *root)
  {
    int noOfLevels, newNoOfLevels;

    newNoOfLevels = RootLS(root);
    if (treeSize == newNoOfLevels)
       return root;

    do {
        noOfLevels = newNoOfLevels;
        root = MinDegOnLevel(noOfLevels);
        newNoOfLevels = RootLS(root);

        if (treeSize == newNoOfLevels)
           return root;

     } while (newNoOfLevels > noOfLevels);
 
    return root;
  }

static void ChangeIndex(PT *p1, PT *p2)
  {
    int index1, index2;

    index1 = p1->indexP;
    index2 = p2->indexP;
    p1->indexP = index2;
    p2->indexP = index1;
    pointOfNumber[index1] = p2;
    pointOfNumber[index2] = p1;

    return;
 }

static int TestNeighbors(PT *p, int level)
  {
    int i, lev, noOfNeigh = NO_OF_PT_NEIGHBORS(p);
    PT **vec=PT_NEIGHBOR_VEC(p);

    for (i = 0; i < noOfNeigh; i++)
      {
        lev =IP(vec[i],levelPt);
        if (lev == level)
            return true;
      }

    return false;
  }

static void NumberThePoint(PT *p)
  {
    PT *pChange;

    pChange = pointOfNumber[lastNo];
    ChangeIndex(p, pChange);
    lastNo--;
    p->mark = false;

    return;
 }

static int Seperate(PT *root)
  {
    PT *p;
    int noOfPoints = actTriang->noOfPoints, sep, no, i, lev;

    sep = 0;    
    root = FindRootND(root);

    levelAt = (levelNo + 1) / 2;


    if (levelAt <= 2)
        for (i=0; i<subGraphNo[treeDepth]; i++)
          {
            p = subGraph[treeDepth][i];
            lev = IP(p,levelPt);
            if ( lev > 0 )
              {
                IP(p,levelPt) = 0;
                NumberThePoint(p);
                sep++;
              }
          }
    else
      {
        treeDepth++;
        subGraph[treeDepth] = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
        if (subGraph[treeDepth] == nil )
          ZIBStdOut("subGraph = nil");

        no = -1;
        for (i=0;i<subGraphNo[treeDepth-1];i++)
          {  
            p = subGraph[treeDepth-1][i];
            lev = IP(p,levelPt);
            if ( lev >= levelAt )
              {
               no++;
               subGraph[treeDepth][no] = p;
               IP(p,treeDepthPt) = treeDepth;

               lev = IP(p,levelPt);
               if ( lev == levelAt )
                 if (TestNeighbors(p, levelAt + 1)) 
                   {
                     IP(p,levelPt) = 0;
                     NumberThePoint(p);
                     sep++;
                   }
               }
          }
          subGraphNo[treeDepth] = no+1;
       }

    return sep;
  }

int NestedDissection()
  {
    int noOfPoints, sep, i;
    PT *p, *root;

    if (treeDepthPt<0)
      treeDepthPt = GetAccess(nodesState->accPoint,sizeof(int),
                              sizeof(int),"treeDepth");
    if (levelPt<0)
      levelPt = GetAccess(nodesState->accPoint,sizeof(int),
                          sizeof(int),"level");
    if ((treeDepthPt<0)||(levelPt<0))
      {
        ZIBStdOut("No access for ints on point in NestedDissection\n");
        return false;
      }
 
    if (!GenPtNeighbor(sortDepth))
         return false;

    noOfPoints = actTriang->noOfPoints;

    pointOfNumber = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (pointOfNumber == nil ) ZIBStdOut("points = nil");
    p = actTriang->firstPoint;
    while (p != nil)
      {
        pointOfNumber[p->indexP] = p;
        p =p->next;
      }

    subGraph = (PT***)ZIBAlloc( (long)noOfPoints*sizeof(PTR**) );
    if (subGraph == nil ) ZIBStdOut("subGraph = nil");
    for (i=0; i<noOfPoints; i++)
        subGraph[i] = nil;


    levelStructNext = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (levelStructNext == nil ) ZIBStdOut("levelStructNext = nil");
    structNextNo = -1;

    levelStructLast = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (levelStructLast == nil ) ZIBStdOut("levelStructLast = nil");
    structLastNo = -1;

    ApplyP(Mark, all);
    sep = 0;

    treeDepth  = 0;
    subGraphNo[0] = noOfPoints;
    subGraph[0] = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (subGraph[0] == nil ) ZIBStdOut("subGraph = nil");

    p = actTriang->firstPoint;
    while (p != nil)
      {
        subGraph[0][p->indexP] = p;
        p =p->next;
      }

    ApplyP(SetTreeDepthZero,all);

    noOfPoints = actTriang->noOfPoints;
    lastNo = noOfPoints - 1;

    
    while (sep < noOfPoints)
      {
        root = nil;
        while (root == nil)
          {
           for (i=0; i<subGraphNo[treeDepth]; i++)
             {
               p = subGraph[treeDepth][i];
               if ( p->mark ) 
                  {
                   root = p;
                   break;
                  }
              }
           if (root==nil) 
             {
              if (subGraph[treeDepth] != nil) ZIBFree((PTR)subGraph[treeDepth]);
              subGraph[treeDepth] = nil;
              treeDepth--;
             }
          }  
        sep += Seperate(root);
      }

   if (pointOfNumber!=nil)   ZIBFree((PTR)pointOfNumber);
   if (levelStructLast!=nil) ZIBFree((PTR)levelStructLast);
   if (levelStructNext!=nil) ZIBFree((PTR)levelStructNext);

    /*  Verlistung der Punkte entsprechend neuer Indizierung   */
    if(!RenumberPts())
       ZIBStdOut(" error in Renumber Points \n");

    ReturnAccess(nodesState->accPoint,treeDepthPt);
    treeDepthPt = -1;
    ReturnAccess(nodesState->accPoint,levelPt);
    levelPt = -1;
    SendAppEvent(TriRenumbered, (ptr)actTriang, sizeof(TRIANGULATION), "NestedDissection");
    return RetPtNeighbor();
  }

static int NumberTheNeighbors(PT *p)
  {
    int i, num, noOfNeigh = NO_OF_PT_NEIGHBORS(p);
    PT *pNeigh, **vec=PT_NEIGHBOR_VEC(p);

    num = 0;
    for (i = 0; i < noOfNeigh; i++) {
        pNeigh = vec[i];
        if (pNeigh->mark) {
            NumberThePoint(pNeigh);
            num++;
        }
    }

    return num;
  }

static int RenumberPts()
  {
    int noOfPoints=actTriang->noOfPoints,i;
    PT *first = actTriang->firstPoint, *p;
    PT **pp;

    pp = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (pp == nil ) ZIBStdOut("pp = nil");

    p = first;
    while (p != nil)
     {
        pp[p->indexP] = p;
        p =p->next;
     }

    actTriang->firstPoint = pp[0];
    p = pp[0];
    p->next = pp[1];
    p->last = nil;
    actTriang->lastPoint  = pp[noOfPoints-1];
    p = pp[noOfPoints-1];
    p->next = nil;
    p->last = pp[noOfPoints-2];
 
    for (i=1; i<noOfPoints-1; i++)
      {
        p = pp[i];
        p->next = pp[i+1];
        p->last = pp[i-1];
      }

    if (pp!=nil) ZIBFree((PTR)pp);


    return true;
  }

int ReverseCuthillMcKee()
  {
    int noOfPoints, num, i;
    PT *p, *root;
    PT *first = actTriang->firstPoint;

    if (levelPt<0)
      levelPt = GetAccess(nodesState->accPoint,sizeof(int),
                          sizeof(int),"level");
    if (levelPt<0)
      {
        ZIBStdOut("No access for ints on point in ReverseCuthillMcKee\n");
        return false;
      }

    if (!GenPtNeighbor(sortDepth))
        return false;

    noOfPoints = actTriang->noOfPoints;
    pointOfNumber = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PT*) );
    if (pointOfNumber == nil ) ZIBStdOut("points = nil");
    p = first;
    while (p != nil)
      {
        pointOfNumber[p->indexP] = p;
        p =p->next;
      }

    levelStructNext = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PTR) );
    if (levelStructNext == nil ) ZIBStdOut("levelStructNext = nil");
    structNextNo = -1;

    levelStructLast = (PT**)ZIBAlloc( (long)noOfPoints*sizeof(PTR) );
    if (levelStructLast == nil ) ZIBStdOut("levelStructLast = nil");
    structLastNo = -1;

    ApplyP(Mark, all);
    noOfPoints = actTriang->noOfPoints;
    if (noOfPoints <= 1)
       return true;
    lastNo = noOfPoints - 1;
    p = first;
    root = FindRootRCM(p);
    NumberThePoint(root);
    num = 1;
    for (i = noOfPoints - 1; i >= 0; i--) 
      {
        p = pointOfNumber[i];
        num += NumberTheNeighbors(p);
        if (num >= noOfPoints)
            break;
      }

    if (pointOfNumber!=nil)   ZIBFree((PTR)pointOfNumber);
    if (levelStructLast!=nil) ZIBFree((PTR)levelStructLast);
    if (levelStructNext!=nil) ZIBFree((PTR)levelStructNext);

    /*  Verlistung der Punkte entsprechend neuer Indizierung   */
     if(!RenumberPts())
      ZIBStdOut(" error in Renumber Points \n") ;

    /*   Sortieren der Nachbarn nach index   */
    ReturnAccess(nodesState->accPoint,levelPt);
    levelPt = -1;
    SendAppEvent(TriRenumbered, (ptr)actTriang, sizeof(TRIANGULATION), "ReverseCuthillMcKee");
    return RetPtNeighbor();
  }
