/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>

#include "zibutil.h"
#include "triang2.h"
#include "problem.h"
#include "assemble.h"

static int CompShapeVals(integData*,void(*)(real,real,int,
  real*,real*,real*,real*,real*,real*));
static int UpdateIData(integData*,int,int,
  void (*)(real,real,int,real*,real*,real*,real*,real*,real*),int);

#define REZSQRT3 0.577350269
 
/*
    Integration points and weights
*/
static real BankIPX[] = { SIXTH, TWOTHIRD, SIXTH };
static real BankIPY[] = { SIXTH, SIXTH, TWOTHIRD };
static real BankIW[] = { SIXTH, SIXTH, SIXTH };

static REAL 	FolkIPX[] = { ZERO, ZERO, ONE };
static REAL 	FolkIPY[] = { ZERO, ONE, ZERO };
static REAL 	FolkIW[] = { SIXTH, SIXTH, SIXTH };

static real linIPX[] = { THIRD };
static real linIPY[] = { THIRD };
static real linIW[]  = { HALF };

static real quadIPX[] = { HALF, HALF, ZERO };
static real quadIPY[] = { ZERO, HALF, HALF };
static real quadIW[] = { SIXTH, SIXTH, SIXTH };

static real userIPX[] = { 3.333333333333333e-01, 4.701420641051151e-01,
                          5.971587178976981e-02, 4.701420641051151e-01,
                          1.012865073234563e-01, 7.974269853530872e-01,
                          1.012865073234563e-01 };
static real userIPY[] = { 3.333333333333333e-01, 4.701420641051151e-01,
                          4.701420641051151e-01, 5.971587178976981e-02, 
                          1.012865073234563e-01, 1.012865073234563e-01,
                          7.974269853530872e-01 };
static real userIW[]  = { 0.1125, 6.619707639425308e-02, 6.619707639425308e-02,
                          6.619707639425308e-02, 6.296959027241358e-02,
                          6.296959027241358e-02, 6.296959027241358e-02 };
static int userNoIP = 7;

static int  noOfLineIP = NO_OF_LINE_IPS;
static real lineIPX[3][NO_OF_LINE_IPS] =
 {
  {0.5+0.5/SQRT3,0.5-0.5/SQRT3},
  {0.0,0.0},
  {0.5-0.5/SQRT3,0.5+0.5/SQRT3}
 };
static real lineIPY[3][NO_OF_LINE_IPS] =
 {
  {0.5-0.5/SQRT3,0.5+0.5/SQRT3},
  {0.5-0.5/SQRT3,0.5+0.5/SQRT3},
  {0.0,0.0}
 };
static real lineIW[3][NO_OF_LINE_IPS] =
 {
  {SQRT2/2.0,SQRT2/2.0},
  {0.5,0.5},
  {0.5,0.5}
 };

/*
    Values of shape functions
    
        phi[1]=1-x-y; phi[2|=x; phi[3]=y;
        qPhi[1]=phi[1](2phi[1]-1);
        qPhi[2]=phi[2](2phi[2]-1);
        qPhi[3]=phi[3](2phi[3]-1);
        qPhi[4]=4phi[2]phi[3];
        qPhi[5]=4phi[1]phi[3];
        qPhi[6]=4phi[1]phi[2];
        
    at the integration points [Axelsson, Barker p.174, 177]
*/

void StdShape(real x,real y,int no,real *f,real *fx,real *fy,
              real *fxx,real *fxy,real *fyy)
  {
    switch (no)
    {
      case 0:               /* 1-x-y */
        *f   = ONE-x-y;
        *fx  = -ONE;
        *fy  = -ONE;
        *fxx = ZERO;
        *fxy = ZERO;
        *fyy = ZERO;
        break;
      case 1:               /* x */
        *f   = x;
        *fx  = ONE;
        *fy  = ZERO;
        *fxx = ZERO;
        *fxy = ZERO;
        *fyy = ZERO;
        break;
      case 2:               /* y */
        *f   = y;
        *fx  = ZERO;
        *fy  = ONE;
        *fxx = ZERO;
        *fxy = ZERO;
        *fyy = ZERO;
        break;
      case 3:               /* 4xy */
        *f   = FOUR*x*y;
        *fx  = FOUR*y;
        *fy  = FOUR*x;
        *fxx = ZERO;
        *fxy = FOUR;
        *fyy = ZERO;
        break;
      case 4:               /* 4y(1-x-y) */
        *f   = FOUR*y*(ONE-x-y);
        *fx  = -FOUR*y;
        *fy  = FOUR*(1-x-TWO*y);
        *fxx = -FOUR;
        *fxy = -FOUR;
        *fyy = -EIGHT;
        break;
      case 5:               /* 4x(1-x-y) */
        *f   = FOUR*x*(ONE-x-y);
        *fx  = FOUR*(1-TWO*x-y);
        *fy  = -FOUR*x;
        *fxx = -EIGHT;
        *fxy = -FOUR;
        *fyy = -FOUR;
        break;
    }
    return;
  }

void StdQShape(real x,real y,int no,real *f,real *fx,real *fy,
               real *fxx,real *fxy,real *fyy)
  {
    switch (no)
    {
      case 0:               /* 2(1-x-y)(1/2-x-y) */ 
        *f   = TWO*(ONE-x-y)*(HALF-x-y);
        *fx  = FOUR*(x+y)-THREE;
        *fy  = FOUR*(x+y)-THREE;
        *fxx = FOUR;
        *fxy = FOUR;
        *fyy = FOUR;
        break;
      case 1:               /* x(2x-1) */
        *f   = x*(TWO*x-ONE);
        *fx  = FOUR*x-ONE;
        *fy  = ZERO;
        *fxx = FOUR;
        *fxy = ZERO;
        *fyy = ZERO;
        break;
      case 2:               /* y(2y-1) */
        *f   = y*(TWO*y-ONE);
        *fx  = ZERO;
        *fy  = FOUR*y-ONE;
        *fxx = ZERO;
        *fxy = ZERO;
        *fyy = FOUR;
        break;
      case 3:               /* 4xy */
        *f   = FOUR*x*y;
        *fx  = FOUR*y;
        *fy  = FOUR*x;
        *fxx = ZERO;
        *fxy = FOUR;
        *fyy = ZERO;
        break;
      case 4:               /* 4y(1-x-y) */
        *f   = FOUR*y*(ONE-x-y);
        *fx  = -FOUR*y;
        *fy  = FOUR*(1-x-TWO*y);
        *fxx = ZERO;
        *fxy = -FOUR;
        *fyy = -EIGHT;
        break;
      case 5:               /* 4x(1-x-y) */
        *f   = FOUR*x*(ONE-x-y);
        *fx  = FOUR*(1-TWO*x-y);
        *fy  = -FOUR*x;
        *fxx = -EIGHT;
        *fxy = -FOUR;
        *fyy = ZERO;
        break;
    }
    return;
  }

void StdCShape(real x,real y,int no,real *f,real *fx,real *fy,
               real *fxx,real *fxy,real *fyy)
  {
    switch (no)
    {
      case 0:                       /* (1-x-y)(3(1-x-y)-1)(3(1-x-y)-2)/2 */    
        *f   = (1-x-y)*(3*(1-x-y)-1)*(3*(1-x-y)-2)*0.5;
        *fx  = (-11+36*x+36*y-27*x*x-54*x*y-27*y*y)*0.5;
        *fy  = (-11+36*x+36*y-27*x*x-54*x*y-27*y*y)*0.5;
        *fxx = 18-27*x-27*y;
        *fxy = 18-27*x-27*y;  
        *fyy = 18-27*x-27*y;
        break;
      case 1:                               /* x(3x-1)(3x-2)/2 */  
        *f   = x*(3*x-1)*(3*x-2)*0.5;
        *fx  = (27*x*x-18*x+2)*0.5;
        *fy  = 0;
        *fxx = 27*x-9;
        *fxy = 0;
        *fyy = 0;
        break;
      case 2:                               /* y(3y-1)(3y-2)/2 */  
        *f   = y*(3*y-1)*(3*y-2)*0.5;
        *fx  = 0;
        *fy  = (27*y*y-18*y+2)*0.5;
        *fxx = 0;
        *fxy = 0;
        *fyy = 27*y-9;
        break;
      case 3:                               /* 9xy(3x-1)/2 */       
        *f   = x*y*(3*x-1)*4.5;
        *fx  = (6*x*y-y)*4.5;
        *fy  = (3*x*x-x)*4.5;
        *fxx = 27*y;
        *fxy = (6*x-1)*4.5;
        *fyy = 0;
        break;
      case 4:                               /* 9xy(3y-1)/2 */        
        *f   = x*y*(3*y-1)*4.5;
        *fx  = (3*y*y-y)*4.5;
        *fy  = (6*x*y-x)*4.5;
        *fxx = 0;
        *fxy = (6*y-1)*4.5;
        *fyy = 27*x;
        break;
      case 5:                               /* 9y(1-x-y)(3y-1)/2 */       
        *f   = y*(1-x-y)*(3*y-1)*4.5;
        *fx  = (y-3*y*y)*4.5;
        *fy  = (-1+x+8*y-6*x*y-9*y*y)*4.5;
        *fxx = 0;
        *fxy = (1-6*y)*4.5;
        *fyy = +36-27*x-81*y;
        break;
      case 6:                               /* 9(1-x-y)y(3(1-x-y)-1)/2 */    
        *f   = (1-x-y)*y*(3*(1-x-y)-1)*4.5;
        *fx  = (-5*y+6*x*y+6*y*y)*4.5;
        *fy  = (2-5*x-10*y+3*x*x+12*x*y+9*y*y)*4.5;
        *fxx = 27*y;
        *fxy = (-5+6*x+12*y)*4.5;
        *fyy = -45+54*x+81*y;
        break;
      case 7:                               /* 9(1-x-y)x(3(1-x-y)-1)/2 */       
        *f   = (1-x-y)*x*(3*(1-x-y)-1)*4.5;
        *fx  = (2-10*x-5*y+9*x*x+12*x*y+3*y*y)*4.5;
        *fy  = (-5*x+6*x*x+6*x*y)*4.5;
        *fxx = -45+54*y+81*x;
        *fxy = (-5+12*x+6*y)*4.5;
        *fyy = 27*x;
        break;
      case 8:                               /* 9x(1-x-y)(3x-1)/2 */        
        *f   = x*(1-x-y)*(3*x-1)*4.5;
        *fx  = (-1+y+8*x-6*x*y-9*x*x)*4.5;
        *fy  = (x-3*x*x)*4.5;
        *fxx = +36-27*y-81*x;
        *fxy = (1-6*x)*4.5;
        *fyy = 0;
        break;
      case 9:                               /* 27xy(1-x-y) */       
        *f   = 27*x*y*(1-x-y);
        *fx  = - 27*y*(2*x+y-1);
        *fy  = - 27*x*(x+2*y-1);
        *fxx = - 54*y;
        *fxy = - 27*(2*x+2*y-1);
        *fyy = - 54*x ;
        break;
    }
    return;
  }

void StdC1Shape(real x,real y,int no,real *f,real *fx,real *fy,
               real *fxx,real *fxy,real *fyy)
  {
    switch (no)
    {
      case 0:       /*(1-x-y)[(1-x+2y)(1+2x-y)-16xy] */    
        *f   = (1-x-y)*((1-x+2*y)*(1+2*x-y)-16*x*y);
        *fx  = 6*x*x+26*x*y-6*x+13*y*y-13*y ;
        *fy  = 13*x*x+26*x*y-13*x+6*y*y-6*y ;
        *fxx = 2*(6*x+13*y-3) ;
        *fxy = 13*(2*x+2*y-1);  
        *fyy = 2*(13*x+ 6*y-3);
        break;
      case 1:                               /* x(1-x-2y)(1-x-y) */  
        *f   = x*(1-x-2*y)*(1-x-y);
        *fx  = 3*x*x+6*x*y-4*x+2*y*y-3*y+1;
        *fy  = x*(3*x+4*y-3) ;
        *fxx = 2*(3*x+3*y-2);
        *fxy = 6*x+4*y-3;
        *fyy = 4*x;
        break;
      case 2:                               /* y(1-2x-y)(1-x-y) */  
        *f   = y*(1-2*x-y)*(1-x-y);
        *fx  = y*(4*x+3*y-3);
        *fy  = 2*x*x+6*x*y-3*x+3*y*y-4*y+1 ;
        *fxx = 4*y;
        *fxy = 4*x+6*y-3;
        *fyy = 2*(3*x+3*y-2);
        break;
      case 7:                               /* xx(3-2x)-7xy(1-x-y) */       
        *f   = x*x*(3-2*x)-7*x*y*(1-x-y);
        *fx  = -(6*x*x-14*x*y-6*x-7*y*y+7*y) ;
        *fy  =  7*x*(x+2*y-1);
        *fxx = -2*(6*x-7*y-3);
        *fxy =  7*(2*x+2*y-1);
        *fyy =  14*x;
        break;
      case 8:                               /* xx(x-1)+2xy(1-x-y) */        
        *f   = x*x*(x-1)+2*x*y*(1-x-y);
        *fx  =  3*x*x-4*x*y-2*x-2*y*y+2*y ;
        *fy  = -2*x*(x+2*y-1);
        *fxx =  2*(3*x-2*y-1);
        *fxy = -2*(2*x+2*y-1);
        *fyy = -4*x;
        break;
      case 3:                               /* -xy(1-2x-y) */       
        *f   = -x*y*(1-2*x-y);
        *fx  = y*(4*x+y-1);
        *fy  = x*(2*x+2*y-1);
        *fxx = 4*y;
        *fxy = 4*x+2*y-1;
        *fyy = 2*x;
        break;
      case 4:                               /* yy(3-2y)-7xy(1-x-y) */       
        *f   = y*y*(3-2*y)-7*x*y*(1-x-y);
        *fx  = 7*y*(2*x+y-1) ;
        *fy  = 7*x*x+14*x*y-7*x-6*y*y+6*y;
        *fxx = 14*y;
        *fxy = 7*(2*x+2*y-1);
        *fyy = 2*(7*x-6*y+3);
        break;
      case 5:                               /* -xy(1-x-2y) */       
        *f   = -x*y*(1-x-2*y);
        *fx  = y*(2*x+2*y-1);
        *fy  = x*(x+4*y-1);
        *fxx = 2*y;
        *fxy = 2*x+4*y-1;
        *fyy = 4*x;
        break;
      case 6:                               /* yy(y-1)+2xy(1-x-y) */        
        *f   = y*y*(y-1)+2*x*y*(1-x-y);
        *fx  = - 2*y*(2*x+y-1);
        *fy  = - (2*x*x+4*x*y-2*x-3*y*y+2*y) ;
        *fxx = - 4*y;
        *fxy = - 2*(2*x+2*y-1);
        *fyy = - 2*(2*x-3*y+1);
        break;
      case 9:                               /* 27xy(1-x-y) */       
        *f   = 27*x*y*(1-x-y);
        *fx  = - 27*y*(2*x+y-1);
        *fy  = - 27*x*(x+2*y-1);
        *fxx = - 54*y;
        *fxy = - 27*(2*x+2*y-1);
        *fyy = - 54*x ;
        break;
    }
    return;
  }

static int CompShapeVals(integData *iData,void (*ShapeF)(real,real,int,
                    real*,real*,real*,real*,real*,real*))
  {
    int noIP = iData->noOfIPoints, noSF = iData->noOfShapeFunc;
    int i, j, k;
    long lng;
    real **buffer, *buf;
    real dum1,dum2,dum3,dum4,dum5;

    lng = 6*((noIP*noSF)*sizeof(real)+noSF*sizeof(real*));

    buffer = (real**) ZIBAlloc(lng);
    if (buffer==nil)
      {
        ZIBStdOut("Assemble: not enough memory (CompShapeVals)\n");
        return false;
      }

    iData->shape   = buffer; buffer += noSF;
    iData->shapeX  = buffer; buffer += noSF;
    iData->shapeY  = buffer; buffer += noSF;
    iData->shapeXX = buffer; buffer += noSF;
    iData->shapeXY = buffer; buffer += noSF;
    iData->shapeYY = buffer; buf = (real*)(buffer+noSF);
    for (k = 0; k<noSF; k++)
      {
        (iData->shape)[k]   = buf; buf += noIP;
        (iData->shapeX)[k]  = buf; buf += noIP;
        (iData->shapeY)[k]  = buf; buf += noIP;
        (iData->shapeXX)[k] = buf; buf += noIP;
        (iData->shapeXY)[k] = buf; buf += noIP;
        (iData->shapeYY)[k] = buf; buf += noIP;
        for (i = 0; i<noIP; i++)
            (*ShapeF)((iData->integPointX)[i],
                      (iData->integPointY)[i],
                      k,
                      &((iData->shape)[k][i]),
                      &((iData->shapeX)[k][i]),
                      &((iData->shapeY)[k][i]),
                      &((iData->shapeXX)[k][i]),
                      &((iData->shapeXY)[k][i]),
                      &((iData->shapeYY)[k][i])
                     );
      }

    for(i=0; i<6; i++)
      {
        for(k=0; k<3; k++)
          {
            for(j=0; j<(iData->noOfLineIP); j++)
              { 
                (*ShapeF)((iData->lineIPX)[k][j],
                          (iData->lineIPY)[k][j],
                          i,
                          &((iData->lineVals)[i][k][j]),
                          &dum1,&dum2,&dum3,&dum4,&dum5
                         );   
              }  
          }    
      }
    return true;
  }

static int UpdateIData(integData *iData,int iFormula,int SF,
                void (*ShapeF)(real,real,int,
                  real*,real*,real*,real*,real*,real*), int symP)
  {
    real ***lineVals;
    int i,k;

    switch (iFormula)
    {
      case BANKIP:
        iData->integPointX = BankIPX;
        iData->integPointY = BankIPY;
        iData->integWeight = BankIW;
        iData->noOfIPoints = 3;
        break;
	case FOLKIP:
		iData->integPointX = FolkIPX;
		iData->integPointY = FolkIPY;
		iData->integWeight = FolkIW;
		iData->noOfIPoints = 3;
		break;
      case LINIP:
        iData->integPointX = linIPX;
        iData->integPointY = linIPY;
        iData->integWeight = linIW;
        iData->noOfIPoints = 1;
        break;
      case QUADIP:
        iData->integPointX = quadIPX;
        iData->integPointY = quadIPY;
        iData->integWeight = quadIW;
        iData->noOfIPoints = 3;
        break;
      case USERIP:
        iData->integPointX = userIPX;
        iData->integPointY = userIPY;
        iData->integWeight = userIW;
        iData->noOfIPoints = userNoIP;
        break;
      default:
        sprintf(globBuf, "Assemble: integration formula %d not known\n",
                iFormula);
        ZIBStdOut(globBuf);
        ZIBFree((PTR)iData);
        return false;
    }
    iData->symP = symP;
    iData->noOfShapeFunc = SF;
    iData->noOfLineIP = noOfLineIP;
    iData->lineIPX = (real**)ZIBAlloc(3*sizeof(real*));
    iData->lineIPY = (real**)ZIBAlloc(3*sizeof(real*));
    iData->lineIW = (real**)ZIBAlloc(3*sizeof(real*));
    for (k=0; k<3; k++)
      {
        (iData->lineIPX)[k] = &(lineIPX[k][0]);
        (iData->lineIPY)[k] = &(lineIPY[k][0]);
        (iData->lineIW)[k] = &(lineIW[k][0]);
      }

    lineVals = (real***)ZIBAlloc(6*sizeof(real**));
    if (lineVals==nil) return false;

    lineVals[0] = (real**)ZIBAlloc(6*3*sizeof(real*));
    if (lineVals[0]==nil) return false;
    for (k=1; k<6; k++) lineVals[k] = &(lineVals[0][k*3]);

    lineVals[0][0] = (real*)ZIBAlloc(6*3*NO_OF_LINE_IPS*sizeof(real));
    if (lineVals[0][0]==nil) return false;
    for (k=0; k<6; k++)
      for (i=0; i<3; i++)
        lineVals[k][i] = &(lineVals[0][0][(k*3+i)*NO_OF_LINE_IPS]);

    iData->lineVals = (real***)lineVals;
    CompShapeVals(iData, ShapeF);
    return true;
  }

integData *NewIData(int iFormula, int SF, void (*ShapeF)(real,real,int,
                  real*,real*,real*,real*,real*,real*), int symP)
  {
    integData *iData;

    iData = (integData*)ZIBAlloc((long)sizeof(integData));
    if (iData==nil)
      { ZIBStdOut("Assemble: not enough memory (NewIData)\n"); return nil; }

    UpdateIData(iData, iFormula, SF, ShapeF, symP);
    return iData;
  }
 
real *ExpVecFull(real *x,int oldLng,int newLng)
  {
    int k;
    real *y = nil;
 
	if (newLng>0)
	  {
		y = (real*)ZIBAlloc((long)newLng*sizeof(real));
		if (y==nil)
		  { ZIBStdOut("Assemble: not enough memory (ExpVecFull)\n"); return nil; }
		for(k = 0; k<oldLng; k++) y[k] = x[k];
	  }
    if (x!=nil) ZIBFree((PTR)x);
    return y;
  }

real **ExpMatFull(real **x,int oldLng,int newLng,int symP)
  {
    int i, k;
    real **y = nil;
 
	if (oldLng<0) oldLng = 0;
	if (newLng>0)
	  {
		y = (real**)ZIBAlloc((long)newLng*sizeof(real*));
		if (y==nil)
		  { ZIBStdOut("Assemble: not enough memory (ExpMatFull)\n");  return nil; }
		for(k = 0; k<newLng; k++)
		  {
			y[k] = (real*)ZIBAlloc((long)((symP==SYMMETRIC)?(k+1):newLng)*sizeof(real));
			if (y[k]==nil)
			  { ZIBStdOut("Assemble: not enough memory (ExpMatFull)\n"); return nil; }
		  }
		for(k = 0; k<oldLng; k++)
		  for (i = 0; i<(symP==SYMMETRIC)?(k+1):oldLng; i++)
		    y[k][i] = x[k][i];
      }
    if (x!=nil)
      {
        for (k=0; k<oldLng; k++) ZIBFree((PTR)x[k]);
        ZIBFree((PTR)x);
      }
	return y;
  }
