/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <math.h>

#include "zibutil.h"
#include "appevents.h"
#include "commands.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"
#include "assemble.h"

static int Direct(dirMethod*,int),Iterate(iteMethod*),Estimate(estiMethod*),
           Refine(refMethod*), Solve(), BreakCond(action);
static void PrintBreakCond();
extern StoreVecToNode(real*,int);

static int UserRefEdge(EDG*);

solve *solveState=nil;
int NewSol = -1, NewErrEsti = -1, NewGlobSol = -1;

dirMethod *actDirector=nil, **dirMethods=nil;
iteMethod *actItor=nil, **iteMethods=nil;
estiMethod *actEstor=nil, **estiMethods=nil;
refMethod *actReftor=nil, **refMethods=nil;
matMulMethod *actMatMul=nil, **matMulMethods=nil;
preCondMethod *actPreCond=nil, **preCondMethods=nil;

static void InitSolveData()
  {
    if (solveState==nil)
      {
        solveState = (solve*)ZIBAlloc((long)sizeof(solve));
        if (solveState==nil) return;
		solveState->Direct = Direct;
		solveState->Iterate = Iterate;
		solveState->Estimate = Estimate;
		solveState->Refine = Refine;
		solveState->Solve = Solve;
		solveState->BreakCond = BreakCond;
		solveState->PrintBreakCond = PrintBreakCond;
		solveState->director = actDirector;
		solveState->itor = actItor;
		solveState->estor = actEstor;
		solveState->reftor = actReftor;
	    solveState->ansatz = noAnsatz;

		solveState->verboseP = true;
		solveState->reqGlobError = 1.0e-4;
		solveState->maxSteps = 5;
		solveState->maxDepth = 50;
		solveState->maxPoints = 1000;
      }
    solveState->lastOp = op_start;

    solveState->dirFail = false;
    solveState->estiFail = false;
    solveState->refFail = false;
    solveState->iteFail = false;
    solveState->breakReason = 0;

	if (solveState->director)
	  {
		((solveState->director)->ReleaseAll)();
	  }
	if (solveState->itor)  (solveState->itor)->eps = 1.0e-8;
	if (solveState->estor)  (solveState->estor)->globError = 1.0;

    return;
  }

int DefDirMethod(char *name, char *renumberName, storeMethod method,
    int onlyPosDef, int onlySym,
    int (*Decomp)(),
    int (*FBSubst)(),
    int (*CheckSol)(),
    int (*Assemble)(),
    int (*Renumber)(),
    int (*PrintLS)(printMode),
    int (*ReleaseAll)())
  {
    int k;
    dirMethod *director;

    if (dirMethods==nil)
      {
        dirMethods = (dirMethod**)ZIBAlloc(MAX_DIR_METHODS*(long)sizeof(dirMethod*));
        if (dirMethods==nil) return false;
      }
    for (k=0; k<MAX_DIR_METHODS; k++)
      if (dirMethods[k]==nil) break;
    if (k==MAX_DIR_METHODS)
      {
        sprintf(globBuf,"Maximal %d direct solvers allowed\n",
                MAX_DIR_METHODS);
        ZIBStdOut(globBuf);
        return false;
      }
    director = (dirMethod*)ZIBAlloc((long)sizeof(dirMethod));
    if (director==nil) return false;

    director->name = name;
    director->verboseP = true;
    director->onlyPosDef = onlyPosDef;
    director->onlySym = onlySym;

    director->method = method;
    director->Decomp = Decomp;
    director->FBSubst = FBSubst;
    director->CheckSol = CheckSol;
    director->Assemble = Assemble;
    director->Renumber = Renumber;
    director->PrintLS = PrintLS;
    director->ReleaseAll = ReleaseAll;
    director->renumberName = renumberName;
    director->stiffSparse = -1;
    director->decompSparse = -1;

    dirMethods[k] = director;
    if (solveState!=nil) solveState->director = director;
    actDirector = director;
    AddCmdPar("seldirect", name);

    return true;
  }

int DefIteMethod(char *name, int needsMatrix, int computesRes,
    int onlyPosDef, int onlySym,
    int (*InitMethod)(iteMethod*),
    int (*Method)(iteMethod*),
    int (*CloseMethod)(iteMethod*))
  {
    int k;
    iteMethod *itor;

    if (iteMethods==nil)
      {
        iteMethods = (iteMethod**)ZIBAlloc(MAX_ITE_METHODS*(long)sizeof(iteMethod*));
        if (iteMethods==nil) return false;
      }
    for (k=0; k<MAX_ITE_METHODS; k++)
      if (iteMethods[k]==nil) break;
    if (k==MAX_ITE_METHODS)
      {
        sprintf(globBuf,"Maximal %d iterative solvers allowed\n",
                MAX_ITE_METHODS);
        ZIBStdOut(globBuf);
        return false;
      }
    itor = (iteMethod*)ZIBAlloc((long)sizeof(iteMethod));
    if (itor==nil) return false;

    itor->name = name;
    itor->needsMatrix = needsMatrix;
    itor->computesRes = computesRes;
    itor->onlyPosDef = onlyPosDef;
    itor->onlySym = onlySym;
    itor->InitMethod = InitMethod;
    itor->Method = Method;
    itor->CloseMethod = CloseMethod;
    itor->matMul = actMatMul;
    itor->preCond = actPreCond;
    itor->minSteps = 1;
    itor->maxSteps = 10;
    itor->eps = 1.0e-8;
    itor->res = RMAX;
    itor->iteFactor = 0.01;
    itor->verboseP = true;
    itor->maxRestarts = 1000;
    itor->maxOrtho = 10;

    iteMethods[k] = itor;
    if (solveState!=nil) solveState->itor = itor;
    actItor = itor;
    AddCmdPar("seliterate", name);

    return true;
  }

int DefEstiMethod(char *name,
    int (*InitMethod)(estiMethod*),
    int (*Method)(estiMethod*),
    int (*CloseMethod)(estiMethod*))
  {
    int k;
    estiMethod *estor;

    if (estiMethods==nil)
      {
        estiMethods = (estiMethod**)ZIBAlloc(MAX_ESTI_METHODS*(long)sizeof(estiMethod*));
        if (estiMethods==nil) return false;
      }
    for (k=0; k<MAX_ESTI_METHODS; k++)
      if (estiMethods[k]==nil) break;
    if (k==MAX_ESTI_METHODS)
      {
        sprintf(globBuf,"Maximal %d error estimators allowed\n",
                MAX_ESTI_METHODS);
        ZIBStdOut(globBuf);
        return false;
      }
    estor = (estiMethod*)ZIBAlloc((long)sizeof(estiMethod));
    if (estor==nil) return false;

    estor->name = name;
    estor->verboseP = true;
    estor->ed_rq = -1;
    estor->ed_diag = -1;
    estor->ed_rhs = -1;
    estor->ed_res = -1;
    estor->InitMethod = InitMethod;
    estor->Method = Method;
    estor->CloseMethod = CloseMethod;
    estor->globError = 1.0;

    estiMethods[k] = estor;
    if (solveState!=nil) solveState->estor = estor;
    actEstor = estor;
    AddCmdPar("selestimate", name);

    return true;
  }

int DefRefMethod(char *name, real sigma, real sFactor,
    int (*InitMethod)(refMethod*),
    int (*Method)(refMethod*),
    int (*CloseMethod)(refMethod*))
  {
    int k;
    refMethod *reftor;

    if (refMethods==nil)
      {
        refMethods = (refMethod**)ZIBAlloc(MAX_REF_METHODS*(long)sizeof(refMethod*));
        if (refMethods==nil) return false;
      }
    for (k=0; k<MAX_REF_METHODS; k++)
      if (refMethods[k]==nil) break;
    if (k==MAX_REF_METHODS)
      {
        sprintf(globBuf,"Maximal %d refinement strategies allowed\n",
                MAX_REF_METHODS);
        ZIBStdOut(globBuf);
        return false;
      }
    reftor = (refMethod*)ZIBAlloc((long)sizeof(refMethod));
    if (reftor==nil) return false;

    reftor->name = name;
    reftor->verboseP = true;
    reftor->InitMethod = InitMethod;
    reftor->Method = Method;
    reftor->CloseMethod = CloseMethod;
    reftor->newFactor = 1.0;
    reftor->sigma = sigma;
    reftor->sFactor = sFactor;

    refMethods[k] = reftor;
    if (solveState!=nil) solveState->reftor = reftor;
    actReftor = reftor;
    AddCmdPar("selrefine", name);

    return true;
  }

int DefMatMulMethod(char *name,int holdsSparseMatrix,
    int (*InitMatMul)(int,real**(*)(TR*),int,real*(*)(TR*)),
    int (*MatMul)(int,int),
    int (*CloseMatMul)())
  {
    int k;
    matMulMethod *matMul;

    if (matMulMethods==nil)
      {
        matMulMethods = (matMulMethod**)ZIBAlloc(MAX_MATMUL_METHODS*(long)sizeof(matMulMethod*));
        if (matMulMethods==nil) return false;
      }
    for (k=0; k<MAX_MATMUL_METHODS; k++)
      if (matMulMethods[k]==nil) break;
    if (k==MAX_MATMUL_METHODS)
      {
        sprintf(globBuf,"Maximal %d mat/mul methods allowed\n",
                MAX_MATMUL_METHODS);
        ZIBStdOut(globBuf);
        return false;
      }
    matMul = (matMulMethod*)ZIBAlloc((long)sizeof(matMulMethod));
    if (matMul==nil) return false;

    matMul->name = name;
    matMul->holdsSparseMatrix = holdsSparseMatrix;
    matMul->rhs = -1;
    matMul->stiff = -1;
    matMul->InitMatMul = InitMatMul;
    matMul->MatMul = MatMul;
    matMul->CloseMatMul = CloseMatMul;

    matMulMethods[k] = matMul;
    if (solveState!=nil)
      if ((solveState->itor)!=nil)
        (solveState->itor)->matMul = matMul;
    if (actItor!=nil) actItor->matMul = matMul;
    actMatMul = matMul;
    AddCmdPar("selmatmul", name);

    return true;
  }


int DefPreCondMethod(char *name,
    int (*InitPreCond)(),
    int (*PreCond)(int,int),
    int (*ClosePreCond)())
  {
    int k;
    preCondMethod *preCond;

    if (preCondMethods==nil)
      {
        preCondMethods = (preCondMethod**)ZIBAlloc(MAX_PRECOND_METHODS*(long)sizeof(preCondMethod*));
        if (preCondMethods==nil) return false;
      }
    for (k=0; k<MAX_PRECOND_METHODS; k++)
      if (preCondMethods[k]==nil) break;
    if (k==MAX_PRECOND_METHODS)
      {
        sprintf(globBuf,"Maximal %d mat/mul methods allowed\n",
                MAX_PRECOND_METHODS);
        ZIBStdOut(globBuf);
        return false;
      }
    preCond = (preCondMethod*)ZIBAlloc((long)sizeof(preCondMethod));
    if (preCond==nil) return false;

    preCond->name = name;
    preCond->InitPreCond = InitPreCond;
    preCond->PreCond = PreCond;
    preCond->ClosePreCond = ClosePreCond;

    preCondMethods[k] = preCond;
    if (solveState!=nil)
      if ((solveState->itor)!=nil)
        (solveState->itor)->preCond = preCond;
    if (actItor!=nil) actItor->preCond = preCond;
    actPreCond = preCond;
    AddCmdPar("selprecond", name);

    return true;
  }
/*
extern int stiffIndex;
static int printCount;
static int PrintFirst(char *v)
  {
    char **vec = NODE_NEIGHBOR_VEC(v);
    real *stiff = NODE_SPARSE_VEC(v,stiffIndex);
    int k, lng = NO_OF_NODE_NEIGHBORS(v);

	if  (printCount<0) exit(2);
	printf("%3d: (%6.10f,%6.10f) rhs=%e, sol=%e\n",
	       IV(v,iIndex),RV(v,rX),RV(v,rY),RV(v,rRhs),RV(v,rSol));
	for (k=0; k<lng; k++)
	  printf("%3d: %4.10e ",IV(vec[k],iIndex),stiff[k]);
	printf("\n");
	printCount--;
	return true;
  }
*/
static int Direct(dirMethod *dir, int sysPrint)
  {
    int verboseP;
    real localDirTime, localAssTime, localRenumTime;

    if (solveState==nil)
      {
        ZIBStdOut("Nothing to Solve\n");
        return false;
      }
    if (dir==nil) dir = solveState->director;
    if ((dir->PrintLS)==nil) sysPrint = false;
    verboseP = dir->verboseP;
    localRenumTime = ZIBSeconds();
    if (dir->Renumber)
      {
        if (!((dir->Renumber)())) return false;
        if (verboseP)
          {
            sprintf(globBuf,"Renumbered: %s\n",dir->renumberName);
            ZIBStdOut(globBuf);
          }
      }
    localRenumTime = ZIBSeconds() - localRenumTime;
    solveState->renumbTime += localRenumTime;

    localAssTime = ZIBSeconds();
    InitAss();
    if (!((dir->Assemble)())) return false;
    if (verboseP)
      {
        sprintf(globBuf,"Assembled dim=%d\n",dir->dim);
        ZIBStdOut(globBuf);
      }
    localAssTime = ZIBSeconds() - localAssTime;
    solveState->assTime += localAssTime;
    if (sysPrint) (dir->PrintLS)(printLinSys);

    localDirTime = ZIBSeconds();
    if (!((dir->Decomp)())) return false;
    if (sysPrint) (dir->PrintLS)(printDecomp);
    if (!((dir->FBSubst)())) return false;
    if (verboseP)
      {
        sprintf(globBuf,"Solved: %s\n",dir->name);
        ZIBStdOut(globBuf);
      }
    if ((dir->solution)!=nil)
      StoreVecToNode(dir->solution,nodesState->rSol);
    dir->level = actTriang->refLevel;
    localDirTime = ZIBSeconds() - localDirTime;
    solveState->dirTime += localDirTime;
    solveState->totalTime += localAssTime+localDirTime+localRenumTime;
    if (sysPrint) (dir->PrintLS)(printSol);
    solveState->lastOp = op_dir;
    SendAppEvent(NewSol, (char*)nil, 0, dir->name);
    if (solveState->verboseP)
      {
        sprintf(globBuf,"Direct    %3d   %3d  %5d\n",
                actTriang->refLevel, actTriang->maxDepth,
                GetNoNodes());
        ZIBStdOut(globBuf);
      }
/* printCount=10;ApplyNode(PrintFirst,all); */
    return true;
  }

static int Iterate(iteMethod *itor)
  {
    real localIteTime, localAssTime;

    if (solveState==nil)
      {
        ZIBStdOut("Nothing to Iterate\n");
        return false;
      }
    if (itor==nil) itor = solveState->itor;

    localAssTime = ZIBSeconds();
    InitAss();
/*    NodeSetBound(nodesState->rSol); */

    localAssTime = ZIBSeconds() - localAssTime;
    solveState->assTime += localAssTime;
    localIteTime = ZIBSeconds();
    if (itor->InitMethod)
      { if (!(itor->InitMethod)(itor)) return false; }

    if (((itor->res)>(itor->eps))||(itor->minSteps>0))
      {
        if (itor->Method)
          { if (!(itor->Method)(itor)) return false; }
      }

    if (itor->CloseMethod)
      { if (!(itor->CloseMethod)(itor)) return false; }

    localIteTime = ZIBSeconds() - localIteTime;
    solveState->iteTime += localIteTime;
    solveState->totalTime += localAssTime+localIteTime;
    if (solveState->verboseP)
      {
        sprintf(globBuf,"Iterate   %3d   %3d  %5d             %3d   %9.3e\n",
                actTriang->refLevel, actTriang->maxDepth,
                GetNoNodes(), itor->iteCount,sqrt(itor->res));
        ZIBStdOut(globBuf);
      }

    solveState->lastOp = op_ite;
    SendAppEvent(NewSol, (char*)nil, 0, itor->name);

	return true;
  }

static int Estimate(estiMethod *estor)
  {
    real localEstiTime;

    localEstiTime = ZIBSeconds();

    if (estor->InitMethod)
      { if (!(estor->InitMethod)(estor)) return false; }
    if (estor->Method)
      { if (!(estor->Method)(estor)) return false; }

    if (estor->verboseP)
      {
        sprintf(globBuf, "Estimate: globError=%9.3e(%e)\n",
                estor->globError,sqrt(estor->globError),actReftor->mBar);
        ZIBStdOut(globBuf);
      }
    if (estor->globError<ZERO)
      {
        ZIBStdOut("Estimate: stiffness matrix maybe not positive definite\n");
        return false;
      }
    if (estor->CloseMethod)
      { if (!(estor->CloseMethod)(estor)) return false; }

    localEstiTime = ZIBSeconds() - localEstiTime;
    solveState->estiTime += localEstiTime;
    solveState->totalTime += localEstiTime;
    if (solveState->verboseP)
      {
        sprintf(globBuf,"Estimate  %3d   %3d  %5d   estiErr=%9.3e\n",
           actTriang->refLevel, actTriang->maxDepth,
           actTriang->noOfPoints,sqrt(estor->globError));
        ZIBStdOut(globBuf);
      }
    solveState->lastOp = op_esti;

	return true;
  }

static int Refine(refMethod *reftor)
  {
    real localRefTime;
    real initNoOfPoints = GetNoNodes();

    localRefTime = ZIBSeconds();

    if (reftor->InitMethod)
      { if (!(reftor->InitMethod)(reftor)) return false; }
    if (reftor->Method)
      { if (!(reftor->Method)(reftor)) return false; }
    if (reftor->CloseMethod)
      { if (!(reftor->CloseMethod)(reftor)) return false; }


    if (initNoOfPoints==(actTriang->noOfPoints))
      ZIBStdOut("Refine: nothing to refine?\n");
    else reftor->newFactor = initNoOfPoints/GetNoNodes();
 
    localRefTime = ZIBSeconds() - localRefTime;
    solveState->refTime += localRefTime;
    solveState->totalTime += localRefTime;
    if (solveState->verboseP)
      {
        sprintf(globBuf,"Refine    %3d   %3d  %5d   mbar=%9.3e\n",
           actTriang->refLevel, actTriang->maxDepth,
           actTriang->noOfPoints,reftor->mBar);
        ZIBStdOut(globBuf);
      }
    solveState->lastOp = op_ref;

    return true;
  }

static int BreakCond(action lastOp)
  {
    if ((solveState->maxSteps)<=(solveState->curStep))
      { solveState->breakReason = 1; return true; }
    switch(lastOp)
      {
    case op_start:
        break;
    case op_dir:
        if (solveState->dirFail)
          { solveState->breakReason = -1; return true; }
    case op_ite:
        if (solveState->iteFail)
          { solveState->breakReason = -2; return true; }
        if ((solveState->maxDepth)<=(actTriang->maxDepth))
          { solveState->breakReason = 2; return true; }
        if ((solveState->maxPoints)<=(actTriang->noOfPoints))
          { solveState->breakReason = 3; return true; }
        break;
    case op_esti:
        if (solveState->estiFail)
          { solveState->breakReason = -3; return true; }
        if ((solveState->reqGlobError)>(actEstor->globError))
          { solveState->breakReason = 4; return true; }
        break;
    case op_ref:
        if (solveState->refFail)
          { solveState->breakReason = -4; return true; }
        break;
      }
    return  false;
  }

static void PrintBreakCond()
  {
    int failed = false;

    ZIBStdOut("Solve: state of solution process\n");
    sprintf(globBuf,"  depth     %5d - %5d\n",
            actTriang->maxDepth, solveState->maxDepth);
    ZIBStdOut(globBuf);
    sprintf(globBuf,"  dimension %5d - %5d\n",
            actTriang->noOfPoints, solveState->maxPoints);
    ZIBStdOut(globBuf);
    sprintf(globBuf,"  global error %9.3e - %9.3e\n",
            sqrt(actEstor->globError),
            sqrt(solveState->reqGlobError));
    ZIBStdOut(globBuf);
    sprintf(globBuf, "  new point factor s %9.3e\n", actReftor->sFactor);
    ZIBStdOut(globBuf);
    sprintf(globBuf,"  min/max steps for Iterate: %d-%d\n",
            actItor->minSteps, actItor->maxSteps);
    ZIBStdOut(globBuf);
    sprintf(globBuf, "  iteration factor %9.3e\n", actItor->iteFactor);
    ZIBStdOut(globBuf);
    sprintf(globBuf, "  last iteration error:%9.3e\n",sqrt(actItor->eps));
    ZIBStdOut(globBuf);

    if ((actDirector->dim)>0)
      {
        sprintf(globBuf, "  direct solution for %d nodes available\n",
                actDirector->dim);
        ZIBStdOut(globBuf);
      }
    else ZIBStdOut("  no direct solution available\n");

    if (solveState->dirFail)
      { ZIBStdOut("  Direct failed\n"); failed = true; }
    if (solveState->estiFail)
      { ZIBStdOut("  Estimate failed\n"); failed = true; }
    if (solveState->refFail)
      { ZIBStdOut("  Refine failed\n"); failed = true; }
    if (solveState->iteFail)
      { ZIBStdOut("  Iterate failed\n"); failed = true; }
    if (!failed) ZIBStdOut("  no failures yet\n");

    return;
  }

static int Solve()
  {
    int initNoOfPoints, lastNoOfPoints, returnP = false,
        saveDirVerbose = actDirector->verboseP,
        saveEstiVerbose = actEstor->verboseP,
        saveRefVerbose = actReftor->verboseP,
        saveIteVerbose = actItor->verboseP;

    actDirector->verboseP = ((solveState->verboseP)==2)?true:false;
    actEstor->verboseP = ((solveState->verboseP)==3)?true:false;
    actReftor->verboseP = ((solveState->verboseP)==4)?true:false;
    actItor->verboseP = ((solveState->verboseP)==5)?true:false;

    solveState->curStep = 0;

    if ((solveState->BreakCond)(solveState->lastOp)) return true;
    if ((solveState->lastOp)==op_start)
      {
        if (!(solveState->Direct)(actDirector,false))
          solveState->dirFail = true;
        (solveState->curStep)++;
        if ((solveState->BreakCond)(solveState->lastOp)) returnP = true;
      }
    while (!returnP)
    {
      initNoOfPoints = actTriang->noOfPoints;
      do
      {
        lastNoOfPoints = actTriang->noOfPoints;

        if ((solveState->lastOp)!=op_esti)
          {
            if (!(solveState->Estimate)(actEstor))
              solveState->estiFail=true;
            if ((solveState->BreakCond)(solveState->lastOp))
              { returnP = true; break; }
          }

        if ((solveState->lastOp)==op_esti)
          {
            if (!(solveState->Refine)(actReftor))
              solveState->refFail=true;
            if ((solveState->BreakCond)(solveState->lastOp))
              { returnP = true; break; }
          }

        if (lastNoOfPoints==(actTriang->noOfPoints)) return false;
      } while (initNoOfPoints*(actReftor->sFactor)>(actTriang->noOfPoints));
      if (returnP) break;

      actItor->eps = (actEstor->globError)*(actItor->iteFactor)*(actReftor->newFactor);
      if (!(solveState->Iterate)(actItor))
        solveState->iteFail=true;
      (solveState->curStep)++;
      if ((solveState->BreakCond)(solveState->lastOp))
        { returnP = true; break; }
    }

    actDirector->verboseP = saveDirVerbose;
    actEstor->verboseP = saveEstiVerbose;
    actReftor->verboseP = saveRefVerbose;
    actItor->verboseP = saveIteVerbose;

    if (returnP)
	  SendAppEvent(NewGlobSol, (char*)nil, 0, solveState->name);

    return returnP;
  }

static int UserRefEdge(EDG *ed)
  {
    real dirVal;
    int k;
    PT *p1 = ed->p1, *p2 = ed->p2, *p = ed->pm;
    char *v1, *v2, *vm;
	static int first=true;

    if ((solveState->ansatz)!=linear)
	  {
		if (first)
		  {
		    ZIBStdOut("UserRefEdge: only for linear ansatz useful\n");
			first = false;
		  }
		return true;
	  }
	for (k=0; k<(nodesState->noOfPointNodes); k++)
      {
        vm = &((p->vec)[(nodesState->startPointNodes)[k]]);
        if (((eqnBounds==nil)&&((ed->boundP)==DIRICHLET))||
		    ((eqnBounds!=nil)&&((ed->boundP)!=INTERIOR)&&
			 (eqnBounds[ed->classB][k]==DIRICHLET)))
          {
			(actProblem->Dirichlet)(p->x,p->y,p->classA,&dirVal,
                                    nodesState->eqnPointNodes[k]);
            RV(vm,rSol) = dirVal;
          }
        else
          { 
	        v1 = &((p1->vec)[(nodesState->startPointNodes)[k]]);
	        v2 = &((p2->vec)[(nodesState->startPointNodes)[k]]);
            RV(vm,rSol) = (RV(v1,rSol)+RV(v2,rSol))*0.5;
/*
			if (k==0)
              if ((solveState->lastOp)==op_esti)
			    {
				  RV(vm,rSol) += RE(ed,actEstor->ed_rq)/RE(ed,actEstor->ed_diag);
				}
*/
          }
      }
    return true;
  }

int SolSetUp(int bytesOnNode, femAnsatz ansatz)
  {
    int dummy;

    if (nodesState==nil) InitNodes();
    if (solveState==nil)
      {
        InitSolveData();
        SetEventProc(RefineEdge,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))UserRefEdge,
                     "Compute estimated sol");
        if (solveState==nil) return false;
        SetEventProc(TriClose,POS_AFTER,(void (*)(ptr, int, char*))nil,
                     (void (*)(ptr, int, char*))InitSolveData,
                     "Reset solve state");
      }
    solveState->director = actDirector;
    solveState->itor = actItor;
    solveState->estor = actEstor;
    solveState->reftor = actReftor;

    if (bytesOnNode<0) GetWS(&dummy,&dummy,&dummy,&bytesOnNode);

    switch (ansatz)
    {
      case linear:
        if (!NodeSetUp(bytesOnNode,1,0,0,theProblem->noOfEquations))
          return false;
        solveState->name = "Standard";
        actIntegData = linIntegData;
        break;      
      case quadratic:
        if (!NodeSetUp(bytesOnNode,1,1,0,theProblem->noOfEquations))
          return false;
        solveState->name = "Quadratic";
        if (quadIntegData==nil)
          { 
            quadIntegData = NewIData(BANKIP, 6, StdQShape, theProblem->symP);
            if (quadIntegData==nil) return false;
          }
        actIntegData = quadIntegData;
        break;
      case cubic:
        if (!NodeSetUp(bytesOnNode,1,2,1,theProblem->noOfEquations))
          return false;
        solveState->name = "Cubic";
        if (cubIntegData==nil)
          { 
            cubIntegData = NewIData(USERIP, 10, StdCShape, theProblem->symP);
            if (cubIntegData==nil) return false;
          }
        actIntegData = cubIntegData;
        break;
      default:
        ZIBStdOut("Unknown FEM-Ansatz\n");
      case noAnsatz:
        break;
    }
    nodesState->bytesOnNode = bytesOnNode;
    if (ansatz!=noAnsatz) solveState->ansatz = ansatz;
    InitAss();

    if (NewSol<0)
      NewSol = DefineEventType("Solution computed",true);
    if (NewSol<0) return false;

    if (NewGlobSol<0)
      NewGlobSol = DefineEventType("Global solution computed",true);
    if (NewGlobSol<0) return false;

    if (NewErrEsti<0)
      NewErrEsti = DefineEventType("Error estimated",true);
    if (NewErrEsti<0) return false;

    return true;
  }
