/*
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "zibutil.h"
#include "commands.h"
#include "params.h"
#include "problem.h"
#include "assemble.h"
#include "compadapt.h"

static char *varName[] = {"x"};
static real *constP = nil, *constG = nil;

    /*      Example: The Model Problem      */


static int StdLaplace(real x, real y, int classA, real *fVals,
                         int equation, int variable)
  {
    if (equation!=variable) return F_IGNORE;
    fVals[0]    =  constP[0];     /*  pX  */
    fVals[1]    =  0.0;           /*  pY  */
    fVals[2]    =  0.0;           /*  pY  */
    fVals[3]    =  constP[0];     /*  pY  */

    return F_CONSTANT;
  }
/*
static int StdHelmholtz(real x, real y, int classA, real *fVal,
                         int equation, int variable)
  {
    return F_IGNORE;
  }

static int StdConvection(real x, real y, int classA, real *fVal,
                         int equation, int variable)
  {
    return F_IGNORE;
  }
*/ 
static int Std0Source(real x, real y, int classA, real *fVal,
                     int equation)
  {
    *fVal = constG[1];
    return F_CONSTANT;
  }
 
static int Std1Source(real x, real y, int classA, real *fVal,
                     int equation)
  {
    *fVal = constG[0];
    return F_CONSTANT;
  }
/* 
static int StdObstacle(real x, real y, int classA, real *fVals,
                       int equation)
  {
    return F_IGNORE;
  }

static int StdCauchy(real x, real y, int classA, real *fVals,
                      int equation)
  {
    fVals[0]    = 0.0;
    fVals[1]    = 0.0;
    return F_CONSTANT;
  }
*/ 
static int StdDirichlet(real x ,real y, int classA, real *fVal,
                      int equation)
  {
    *fVal = 0.0;
    return F_CONSTANT ;
  }


/*  Example: Randy Bank */

static real arccos(real x)
  {
    real y=(fabs(x)<1.0e-10)?REALPI2:atan(sqrt(ONE-x*x)/x);
    if (x<=-1.0e-10) y = REALPI+y;
    return y;
  }
 
static int BankDirichlet(real x, real y, int classA, real *fVal,
                   int equation)
  {  
    real r = sqrt(x*x+y*y), phi;

    if (classA==1)
      *fVal = sqrt(sqrt(r));
    else
      {
        phi = (r>0)?arccos(x/r):ZERO;
        if (y<ZERO) phi = TWO*REALPI-phi;
        r = sqrt(sqrt(r));
        *fVal = r*sin(QUARTER*phi);
      }

    return F_VARIABLE;
  }

static int BankSol(real x, real y, int classA, real *fVal,
                   int equation)
  {
    real r = sqrt(x*x+y*y), phi;

    if ((classA==2) || (classA==1))
      *fVal=sqrt(sqrt(r));
    else
      {
        phi = (r>0)?arccos(x/r):ZERO;
        if (y<ZERO) phi = TWO*REALPI-phi;
        r =sqrt(sqrt(r));
        *fVal= r*sin(QUARTER*phi);
      }

    return F_VARIABLE;    
  }

/*  Example: Discontinuous Dirichlet Boundary Conditions    */
 
static int RandUnDirichlet(real x, real y, int classA, real *fVal,
                           int equation)
  {
    *fVal = 0.0;
    if ((x<HALF)&&(y<HALF))   *fVal = ONE;
    if ((x==ZERO)&&(y==HALF)) *fVal = HALF;
    if ((y==ZERO)&&(x==HALF)) *fVal = HALF;

    return F_VARIABLE ;
  }

/*  Example: Discontinuous Coefficients */

static int RegLaplace(real x, real y, int classA, real *fVals,
                        int equation, int variable)
  {
                    /*  pX  */
    fVals[3] = fVals[0] = (classA==0)?constP[1]:constP[0];
    fVals[1] = fVals[2] = ZERO;

    return F_CONSTANT;
  }
 

/*      Example: mixed boundary values          */

static int KreisSource(real x, real y, int classA, real *fVal,
                       int equation)
  {
    *fVal = constG[2];       /*  g   */
    return F_CONSTANT;
  }

static int KreisCauchy(real x, real y, int classA, real *fVals,
                       int equation, int variable)
  {
    real r;
    r = sqrt(x*x + y*y);
    fVals[0] = 0.0;     /*  Sigma   */
    fVals[1] = 2.0*r;       /*  Xi  */

    return F_VARIABLE;   
  }


static int KreisDirich(real x, real y, int classA, real *fVal,
                        int equation)
  {
    *fVal = x*x+y*y;
    return F_VARIABLE;
  }

static int KreisSol(real x, real y, int classA, real *fVals,
                    int equation)
  {
    fVals[0] = x*x+y*y;    
    fVals[1] = 2.0*x;    
    fVals[2] = 2.0*y;    
    return F_VARIABLE;    
  }

/********************************************************/
/*  Problem-Command: Select a special example   */
/********************************************************/

int SetStdProblems()
  {
	if (constP == nil)
	  {
		constP = (real*) NewParamList(nil,"constP",2,sizeof(real),
		                         (char **)nil,T_REAL,0,(char **)nil,
								 (ptr)nil,(int (*)(char*,char*,int))nil,
								 (int (*)(char*))nil);
		constP[0] = 1.0;
		constP[1] = 1.0e6 ;
	  }
	if (constG == nil)
	  {
		constG = (real*) NewParamList(nil,"constG",3,sizeof(real),
		                         (char **)nil,T_REAL,0,(char **)nil,
								 (ptr)nil,(int (*)(char*,char*,int))nil,
								 (int (*)(char*))nil);
		constG[0] = -1.0;
		constG[1] = 0.0;
		constG[2] = -4.0;
	  }
	if (!InitConstAss(constP,constG)) return true;
    if (!SetProblem("bank",varName,
                    StdLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    Std0Source,
                    (int(*)(real,real,int,real*,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    BankDirichlet,
                    BankSol,
                    ConstAss,
                    ConstAssR)) return false;
    if (!SetProblem("banknum",varName,
                    StdLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    Std0Source,
                    (int(*)(real,real,int,real*,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    BankDirichlet,
                    BankSol,
                    NumAss,
                    NumAssR)) return false;
    if (!SetProblem("bounddisc",varName,
                    StdLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    Std0Source,
                    (int(*)(real,real,int,real*,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    RandUnDirichlet,
                    (int(*)(real,real,int,real*,int))nil,
                    ConstAss,
                    ConstAssR)) return false;
    if (!SetProblem("region",varName,
                    RegLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    Std1Source,
                    (int(*)(real,real,int,real*,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    StdDirichlet,
                    (int(*)(real,real,int,real*,int))nil,
                    ConstAss,
                    ConstAssR)) return false;
    if (!SetProblem("circle",varName,
                    StdLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    KreisSource,
                    (int(*)(real,real,int,real*,int))nil,
                    KreisCauchy,
                    KreisDirich,
                    KreisSol,
                    NumAss,
                    NumAssR)) return false;
    if (!SetProblem("stdnum",varName,
                    StdLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    Std1Source,
                    (int(*)(real,real,int,real*,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    StdDirichlet,
                    (int(*)(real,real,int,real*,int))nil,
                    NumAss,
                    NumAssR)) return false;
    if (!SetProblem("std",varName,
                    StdLaplace,
                    (int(*)(real,real,int,real*,int,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    Std1Source,
                    (int(*)(real,real,int,real*,int))nil,
                    (int(*)(real,real,int,real*,int,int))nil,
                    StdDirichlet,
                    (int(*)(real,real,int,real*,int))nil,
                    ConstAss,
                    ConstAssR)) return false;
    return true;
  }
