 /*
C*    Begin Prologue ELLKASK
C   ______________________________________________________________
C
C*  Title
C
C     Adaptive hierarchical finite element code for solving linear
C     scalar second-order 2-D elliptic partial differential equations
C     on general domains
C
C*  Written by        R. Roitzsch, B. Erdmann
C                     with contributions by R. Kornhuber, F. Bornemann
C*  Purpose           Adaptive finite element solution of elliptic partial
C                     differential equations
C*  Method            finite element discretization, conjugate gradients,
C                     hierarchical basis as preconditioner, adaptive  
C                     multi level scheme 
C*  Category          ? - elliptic partial differential equations
C*  Keywords          ellipitc partial differential equations,
C                     finite element method, multi level methods,
C                     adaptive refinement
C*  Version           1.2
C*  Revision          January 1991
C*  Latest Change     January 1991
C*  Library           CodeLib
C*  Code              C
C*  Environment       Unix systems, Sun (graphics),
C                     Macintosh.
C*  Copyright     (c) Konrad Zuse Zentrum fuer
C                     Informationstechnik Berlin
C                     Heilbronner Str. 10, D-1000 Berlin 31
C                     phone 0049+30+89604-0, 
C                     telefax 0049+30+89604-125
C*  Contact           Rainer Roitzsch 
C                     ZIB, Numerical Software Development 
C                     phone: 0049+30+89604-215
C                     e-mail: 
C                     RFC822 notation: roitzsch@sc.zib-berlin.de
C                     X.400: C=de;A=dbp;P=zib-berlin;OU=sc;S=Roitzsch
C
C*  References 
C
C     /1/ P. Deuflhard, P. Leinen, H. Yserentant:
C         Concept of an Adaptive Hierarchical Finite Element Code.
C         IMPACT I, 3-35 (1989)
C
C     /2/ R. Roitzsch:
C         KASKADE User's Manual
C         KASKADE Programmer's Manual
C         ZIB Technical Reports TR-89-4/5
C
C     /3/ Bodo Erdmann, Rainer Roitzsch
C         KASKADE - Numerical Experiments
C         ZIB Technical Report - in preparation
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
C
C*    More:
C     ======
C
C     see /2,3/, ellkask.doc
C
C*    End Prologue ELLKASK

 */

#ifdef STACKSIZE
#include <Memory.h>
#endif
#include <stdio.h>

#include "start.h"
#include "zibutil.h"
#include "fixedsize.h"
#include "commands.h"
#include "triang2.h"
#include "nodes.h"
#include "problem.h"
#include "solve.h"
#include "assemble.h"
#include "graphic.h"
#include "compadapt.h"

extern int RdTri(command*), WrTri(command*), SReadCmd(command*),
           DumpCmd(command*), TecPlotCmd(command*);

/* #define PERFORMANCE */

#ifdef PERFORMANCE
#undef true
#undef false
#include <Perf.h>
#define ROMII 0x78
#endif

#ifdef PERFORMANCE
TP2PerfGlobals ThePGlobals;
OSErr err;
#endif

static char *cmdDefFileName = "ellkask.def";

static void SetAddresses();

char *kaskPath = "";
char *globBuf = nil;

static REAL totalTime, lastTime;

int main(int argc, char **argv, char **envp)
  {
	char *initDoFileName = nil;

#ifdef PERFORMANCE
ThePGlobals = nil;
if (!InitPerf(&ThePGlobals, 4, 8, false, true, "\pCODE",
              0, "", false, 0, 0, 8))
  { ZIBStdOut("Init of performance tools failed\n"); return -1; }
(void)PerfControl(ThePGlobals, true);
#endif

#ifdef MALLOCDEBUG
malloc_debug(2);
#endif

#ifdef STACKSIZE
printf("GetApplLimit()=%ld, STACKSIZE=%ld,new=%ld\n",GetApplLimit(),
(long)STACKSIZE*1024,GetApplLimit()-(long)STACKSIZE);
    SetApplLimit(GetApplLimit()-(long)STACKSIZE*1024);
#endif

    globBuf = (char*)ZIBAlloc(256L);
    if (globBuf==nil) { ZIBStdOut("Aua\n"); return -1; }
    totalTime = ZIBSeconds();
    lastTime = totalTime;

	Z_Unix(argc,argv,envp);
	Z_ModSet(d_mainModulID,d_ModVersion,"2.1");
	Z_ModSet(d_mainModulID,d_PathVarName,"KASKINPUTS");
	Z_Start(d_mainModulID);
	Z_ModGet(d_mainModulID,"do",&initDoFileName);

    if (!InitCommands(kaskPath, cmdDefFileName,
	                  initDoFileName?initDoFileName:"ellkask.startup"))
      { ZIBStdOut("Jammer\n"); return -1; }
    InitFixedLists(100,(void (*)(int, char*))nil);
    SetLibAddresses(-1);
    InitTriang();
    SetTriAddresses(-1);
    SetGraphAddresses();
    SetProblemAddresses();
    SetSolveAddresses();
    InitProblem("ellkask",100,1,100,FULL);
    SetStdProblems();
    SetAddresses();
    DefFullChol();
    DefEnvChol();
    DefSSORMethod();
    DefGMResMethod();
    DefCGSMethod();
    DefBiCGStabMethod();
    DefCGMethod();
    DefMatMulMethod("local",false,InitMatMul,AXMul,(int(*)())nil);
    DefMatMulMethod("localsave",false,InitSaveMul,AXMul,(int(*)())nil);
    DefMatMulMethod("sparse",true,InitSparseMul,AXSparseMul,(int(*)())nil);
    DefNonePreCond(); DefDiagPreCond(); DefBPXPreCond(); DefHBPreCond();
    DefEllErrMethod();

    CmdMainLoop();

    CloseGraphic(-1);
    totalTime = ZIBSeconds()-totalTime;

	Z_End(d_mainModulID,0);

#ifdef PERFORMANCE
err = PerfDump(ThePGlobals, "\pkaskade.out", false, 0);
if (err!= noErr) ZIBStdOut("Writing performance data failed\n");
TermPerf(ThePGlobals);
#endif

    return 0;
  }

static int PrTime(COMMAND *cmd)
  {
    REAL tt = ZIBSeconds();

    if (ParsCheck(cmd,0,0)) return false;

    sprintf(globBuf, "  Time: %.2f secs, since last call %.2f secs, \n",
            tt-totalTime, tt-lastTime);
    ZIBStdOut(globBuf);
    if (solveState!=nil)
      {
        if ((solveState->dirTime)!=ZERO)
          {
            sprintf(globBuf,"    acc. direct time:  %.3f secs\n",
                    solveState->dirTime);
            ZIBStdOut(globBuf);
          }
        if ((solveState->renumbTime)!=ZERO)
          {
            sprintf(globBuf,"    acc. renumb. time: %.3f secs\n",
                    solveState->renumbTime);
            ZIBStdOut(globBuf);
          }
        if ((solveState->assTime)!=ZERO)
          {
            sprintf(globBuf,"    acc. assemb. time: %.3f secs\n",
                    solveState->assTime);
            ZIBStdOut(globBuf);
          }
        if ((solveState->iteTime)!=ZERO)
          {
            sprintf(globBuf,"    acc. iterat. time: %.3f secs\n",
                    solveState->iteTime);
            ZIBStdOut(globBuf);
          }
        if ((solveState->estiTime)!=ZERO)
          {
            sprintf(globBuf,"    acc. estimat. time: %.3f secs\n",
                    solveState->estiTime);
            ZIBStdOut(globBuf);
          }
        if ((solveState->refTime)!=ZERO)
          {
            sprintf(globBuf,"    acc. ref. time: %.3f secs\n",
                    solveState->refTime);
            ZIBStdOut(globBuf);
          }
      }
    sprintf(globBuf,"  Allocated memory: %dbyte\n",allocMem);
    ZIBStdOut(globBuf);

    lastTime = tt;
    return true;
  }

static void SetAddresses()
  {
    SetCommand(10, RdTri);
    SetCommand(11, WrTri);
    SetCommand(18, PrTime);
    SetCommand(28, DumpCmd);
    SetCommand(29, TecPlotCmd);
    SetCommand(51, InfCompAdaptCmd);
    SetCommand(52, InfGeoCmd);
    SetCommand(53, SReadCmd);

    return;
  }
