 /*
C*    Begin Prologue ELLKASK - 3D
C   ______________________________________________________________
C
C*  Title
C
C     Adaptive hierarchical finite element code for solving linear
C     scalar second-order 3-D elliptic partial differential equations
C     on general domains
C
C*  Written by        R. Roitzsch, B. Erdmann
C                     with contributions by R. Kornhuber, F. Bornemann
C*  Purpose           Adaptive finite element solution of elliptic partial
C                     differential equations
C*  Method            finite element discretization, conjugate gradients,
C                     hierarchical basis as preconditioner, adaptive  
C                     multi level scheme 
C*  Category          ? - elliptic partial differential equations
C*  Keywords          ellipitc partial differential equations,
C                     finite element method, multi level methods,
C                     adaptive refinement
C*  Version           2.0
C*  Revision          January 1993
C*  Latest Change     January 1993
C*  Library           CodeLib
C*  Code              C
C*  Environment       Unix systems, Sun (graphics),
C                     Macintosh.
C*  Copyright     (c) Konrad Zuse Zentrum fuer
C                     Informationstechnik Berlin
C                     Heilbronner Str. 10, D-1000 Berlin 31
C                     phone 0049+30+89604-0, 
C                     telefax 0049+30+89604-125
C*  Contact           Bodo Erdmann 
C                     ZIB, Numerical Software Development 
C                     phone: 0049+30+89604-215
C                     e-mail: 
C                     RFC822 notation: erdmann@sc.zib-berlin.de
C                     X.400: C=de;A=dbp;P=zib-berlin;OU=sc;S=Erdmann
C
C*  References 
C
C     /1/ P. Deuflhard, P. Leinen, H. Yserentant:
C         Concept of an Adaptive Hierarchical Finite Element Code.
C         IMPACT I, 3-35 (1989)
C
C     /2/ F. Bornemann, B. Erdmann, R. Kornhuber:
C         Adaptive Multilevel-Methods in Three Space Dimensions.
C         ZIB Preprint SC 92-14 (1992)
C
C     /3/ R. Roitzsch:
C         KASKADE Programmer's Manual
C         ZIB Technical Reports, to appear
C
C     /4/ F. Bornemann, B. Erdmann, R. Roitzsch
C         KASKADE - Numerical Experiments
C         ZIB Technical Reports TR-91-1 (1991)
C
C  _______________________________________________________________
C
C*   Licence
C    =======
C
C    You may use or modify this code for your own non commercial
C    purposes for an unlimited time. 
C    In any case you should not deliver this code without a special 
C    permission of ZIB.
C    In case you intend to use the code commercially, we oblige you
C    to sign an according licence agreement with ZIB.
C
C
C  _______________________________________________________________
C
C
C*    More:
C     ======
C
C     see /2,3/, ellkask3d.doc
C
C*    End Prologue ELLKASK - 3D

 */
#include <stdio.h>

#include "start.h"
#include "zibutil.h"
#include "fixedsize.h"
#include "commands.h"
#include "triang3.h"
#include "nodes3.h"
#include "problem3.h"
#include "solve3.h"
#include "assemble3.h"
#include "graphic3.h"

extern int InitGauss();
extern int RdTri(COMMAND*), RdTriHyper(COMMAND*);
extern int WrTri(COMMAND*);
extern int Grape(COMMAND*), GrapeClass(COMMAND*);
extern int WriteInt(COMMAND*);

/* #define PERFORMANCE */

#ifdef PERFORMANCE
#undef true
#undef false
#include <Perf.h>
#define ROMII 0x78
#endif

#ifdef PERFORMANCE
TP2PerfGlobals ThePGlobals;
OSErr err;
#endif

static char *cmdDefFileName = "ellkask3d.def";

static void SetAddresses();

char *kaskPath = "";
char *globBuf = nil;

static REAL startTime, totalTime, lastTime;

int main(int argc, char **argv, char **envp)   
  {
	char *initDoFileName = nil;

#ifdef PERFORMANCE
ThePGlobals = nil;
if (!InitPerf(&ThePGlobals, 4, 8, false, true, "\pCODE",
			  ROMII, "", false, 0, 0, 8))
  { ZIBStdOut("Init of performance tools failed\n"); return -1; }
(void)PerfControl(ThePGlobals, true);
#endif

#ifdef MALLOCDEBUG
malloc_debug(2);
#endif

	globBuf = (char*)ZIBAlloc(256L);
	if (globBuf==nil) { ZIBStdOut("Aua\n"); return -1; }
	startTime = ZIBSeconds();
	lastTime = startTime;

	Z_Unix(argc,argv,envp);
	Z_ModSet(d_mainModulID,d_ModVersion,"2.1");
	Z_ModSet(d_mainModulID,d_PathVarName,"KASKINPUTS");
	Z_Start(d_mainModulID);
	Z_ModGet(d_mainModulID,"do",&initDoFileName);

    if (!InitCommands(kaskPath, cmdDefFileName,
	                  initDoFileName?initDoFileName:"ellkask.startup"))
      { ZIBStdOut("Jammer\n"); return -1; }

	InitFixedLists(100,(void (*)(int, char*))nil);
	SetLibAddresses(-1);
	InitTriang();
	SetTriAddresses(-1);
	SetGraphAddresses();
	SetProblemAddresses();
	SetSolveAddresses();
	InitProblem("ellkask3d",100,1,100,SYMMETRIC);
	SetStdProblems();
	SetAddresses();
	DefFullChol();
	DefEnvChol();
	DefSSORMethod();
	DefBiCGStabMethod();
	DefCGMethod();
	DefMatMulMethod("local",false,InitMatMul,AXMul,(int(*)())nil);
        DefMatMulMethod("localsave",false,InitSaveMul,AXMul,(int(*)())nil);
	DefMatMulMethod("sparse",true,InitSparseMul,AXSparseMul,(int(*)())nil);
	DefNonePreCond(); DefDiagPreCond(); DefHBPreCond(); DefBPXPreCond();
	DefEllErrMethod();
	CmdMainLoop();

	CloseGraphic(-1);
	totalTime = ZIBSeconds()-startTime;

	Z_End(d_mainModulID,0);

#ifdef PERFORMANCE
err = PerfDump(ThePGlobals, "\pkaskade.out", false, 0);
if (err!= noErr) ZIBStdOut("Writing performance data failed\n");
TermPerf(ThePGlobals);
#endif

	return 0;
  }

static int PrTime(COMMAND *cmd)
{
   REAL tt = ZIBSeconds();
   REAL dt = tt - startTime;
   REAL t;

   if (ParsCheck(cmd,0,0)) return false;

   sprintf(globBuf, "  Time: %.2f secs, since last call %.2f secs, \n",
            dt, tt-lastTime);
   ZIBStdOut(globBuf);
   if (solveState!=nil)
     {
      t = solveState->totalTime;
      if ((solveState->dirTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. direct time:        %6.3f secs (%4.2f %%)\n",
             solveState->dirTime, 100*solveState->dirTime/t);
          ZIBStdOut(globBuf);
         }
      if ((solveState->renumbTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. renumb. time:       %6.3f secs (%4.2f %%)\n",
             solveState->renumbTime, 100*solveState->renumbTime/t);
            ZIBStdOut(globBuf);
        }
      if ((solveState->assTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. assemble time:      %6.3f secs (%4.2f %%)\n",
             solveState->assTime, 100*solveState->assTime/t);
          ZIBStdOut(globBuf);
        }
      if ((solveState->iteTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. iterat. time:       %6.3f secs (%4.2f %%)\n",
                  solveState->iteTime,100*solveState->iteTime/t);
          ZIBStdOut(globBuf);
        }
      if ((solveState->estiTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. estimat. time:      %6.3f secs (%4.2f %%)\n",
                  solveState->estiTime,100*solveState->estiTime/t);
          ZIBStdOut(globBuf);
          sprintf(globBuf,"          Note: estim.time might include time for \n");
          ZIBStdOut(globBuf);
          sprintf(globBuf,"          computing exact error and norm of approx.sol\n");
          ZIBStdOut(globBuf);
        }
      if ((solveState->refTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. ref. time:          %6.3f secs (%4.2f %%)\n",
                  solveState->refTime,100*solveState->refTime/t);
          ZIBStdOut(globBuf);
        }
      if ((solveState->totalTime)!=ZERO)
        {
          sprintf(globBuf,"    acc. time:               %6.3f secs\n", t);
          ZIBStdOut(globBuf);
        }
      }
    sprintf(globBuf,"  Allocated memory: %dbyte\n",allocMem);
    ZIBStdOut(globBuf);

    lastTime = tt;
    return true;
  }

static void SetAddresses()
  {
	SetCommand(10, RdTri);
	SetCommand(20, RdTriHyper);
	SetCommand(11, WrTri);
	SetCommand(25, Grape);
	SetCommand(26, GrapeClass);
	SetCommand(18, PrTime);
    SetCommand(51, WriteInt);

	return;
  }
